import torch

from torchpack.datasets.dataset import Dataset
from torchvision import datasets, transforms
from typing import List
from torchpack.utils.logging import logger
from torchvision.transforms import InterpolationMode


__all__ = ["CIFAR10"]


resize_modes = {
    "bilinear": InterpolationMode.BILINEAR,
    "bicubic": InterpolationMode.BICUBIC,
    "nearest": InterpolationMode.NEAREST,
}


class CIFAR10Dataset:
    def __init__(
        self,
        root: str,
        split: str,
        train_valid_split_ratio: List[float],
        center_crop,
        resize,
        resize_mode,
        binarize,
        binarize_threshold,
        grayscale,
        digits_of_interest,
        n_test_samples,
        n_valid_samples,
        fashion,
    ):
        self.root = root
        self.split = split
        self.train_valid_split_ratio = train_valid_split_ratio
        self.data = None
        self.center_crop = center_crop
        self.resize = resize
        self.resize_mode = resize_modes[resize_mode]
        self.binarize = binarize
        self.binarize_threshold = binarize_threshold
        self.grayscale = grayscale
        self.digits_of_interest = digits_of_interest
        self.n_test_samples = n_test_samples
        self.n_valid_samples = n_valid_samples
        self.fashion = fashion

        self.load()
        self.n_instance = len(self.data)

    def load(self):
        if self.grayscale:
            tran = [
                transforms.ToTensor(),
                transforms.Grayscale(num_output_channels=1),
                transforms.Normalize(
                    (0.2989 * 0.4914 + 0.587 * 0.4822 + 0.114 * 0.4465,),
                    (
                        (
                            (0.2989 * 0.2023) ** 2
                            + (0.587 * 0.1994) ** 2
                            + (0.114 * 0.2010) ** 2
                        )
                        ** 0.5,
                    ),
                ),
            ]
        else:
            tran = [
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                ),
            ]

        if not self.center_crop == 32:
            tran.append(transforms.CenterCrop(self.center_crop))
        if not self.resize == 32:
            tran.append(transforms.Resize(self.resize, interpolation=self.resize_mode))
        transform = transforms.Compose(tran)

        if self.split == "train" or self.split == "valid":
            train_valid = datasets.CIFAR10(
                self.root, train=True, download=True, transform=transform
            )
            targets = torch.tensor(train_valid.targets)
            idx, _ = torch.stack(
                [targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            # targets = targets[idx]
            train_valid.targets = targets[idx].numpy().tolist()
            train_valid.data = train_valid.data[idx]

            train_len = int(self.train_valid_split_ratio[0] * len(train_valid))
            split = [train_len, len(train_valid) - train_len]
            train_subset, valid_subset = torch.utils.data.random_split(
                train_valid, split, generator=torch.Generator().manual_seed(1)
            )
            if self.split == "train":
                self.data = train_subset
            else:
                if self.n_valid_samples is None:
                    # use all samples in valid set
                    self.data = valid_subset
                else:
                    # use a subset of valid set, useful to speedup evo search
                    valid_subset.indices = valid_subset.indices[: self.n_valid_samples]
                    self.data = valid_subset
                    logger.warning(
                        f"Only use the front "
                        f"{self.n_valid_samples} images as "
                        f"VALID set."
                    )

        else:
            test = datasets.CIFAR10(self.root, train=False, transform=transform)
            targets = torch.tensor(test.targets)
            idx, _ = torch.stack(
                [targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            test.targets = targets[idx].numpy().tolist()
            test.data = test.data[idx]
            if self.n_test_samples is None:
                # use all samples as test set
                self.data = test
            else:
                # use a subset as test set
                test.targets = test.targets[: self.n_test_samples]
                test.data = test.data[: self.n_test_samples]
                self.data = test
                logger.warning(
                    f"Only use the front {self.n_test_samples} " f"images as TEST set."
                )

    def __getitem__(self, index: int):
        img = self.data[index][0]
        if self.binarize:
            img = 1.0 * (img > self.binarize_threshold) + -1.0 * (
                img <= self.binarize_threshold
            )

        digit = self.digits_of_interest.index(self.data[index][1])
        instance = {"image": img, "digit": digit}
        return instance

    def __len__(self) -> int:
        return self.n_instance


class CIFAR10(Dataset):
    def __init__(
        self,
        root: str,
        train_valid_split_ratio: List[float],
        center_crop=32,
        resize=32,
        resize_mode="bilinear",
        binarize=False,
        binarize_threshold=0.1307,
        grayscale=False,
        digits_of_interest=tuple(range(10)),
        n_test_samples=None,
        n_valid_samples=None,
        fashion=False,
    ):
        self.root = root

        super().__init__(
            {
                split: CIFAR10Dataset(
                    root=root,
                    split=split,
                    train_valid_split_ratio=train_valid_split_ratio,
                    center_crop=center_crop,
                    resize=resize,
                    resize_mode=resize_mode,
                    binarize=binarize,
                    binarize_threshold=binarize_threshold,
                    grayscale=grayscale,
                    digits_of_interest=digits_of_interest,
                    n_test_samples=n_test_samples,
                    n_valid_samples=n_valid_samples,
                    fashion=fashion,
                )
                for split in ["train", "valid", "test"]
            }
        )


if __name__ == "__main__":
    import pdb

    pdb.set_trace()
    cifar10 = CIFAR10Dataset(
        root="../cifar10_data",
        split="train",
        train_valid_split_ratio=[0.9, 0.1],
        center_crop=32,
        resize=32,
        resize_mode="bilinear",
        binarize=False,
        binarize_threshold=0.1307,
        grayscale=True,
        digits_of_interest=(3, 6),
        n_test_samples=100,
        n_valid_samples=1000,
        fashion=True,
    )
    cifar10.__getitem__(20)
    print("finish")



import torch
import torchquantum as tq
import pathos.multiprocessing as multiprocessing
import itertools

from qiskit import Aer, execute, IBMQ, transpile, QuantumCircuit
from qiskit.providers.aer.noise import NoiseModel
from qiskit.tools.monitor import job_monitor
from qiskit.exceptions import QiskitError
from torchquantum.plugins import (
    tq2qiskit,
    tq2qiskit_parameterized,
    tq2qiskit_measurement,
)
from torchquantum.utils import (
    get_expectations_from_counts,
    get_provider,
    get_provider_hub_group_project,
    get_circ_stats,
)
from .qiskit_macros import IBMQ_NAMES
from tqdm import tqdm
from torchpack.utils.logging import logger
from qiskit.transpiler import PassManager
import numpy as np
import datetime

from torchquantum.plugins import my_job_monitor


class EmptyPassManager(PassManager):
    def run(self, circuits, output_name: str = None, callback=None):
        return circuits


def run_job_worker(data):
    while True:
        try:
            job = execute(**(data[0]))
            qiskit_verbose = data[1]
            if qiskit_verbose:
                job_monitor(job, interval=1)
            result = job.result()
            counts = result.get_counts()
            # circ_num = len(data[0]['parameter_binds'])
            # logger.info(
            #     f'run job worker successful, circuit number = {circ_num}')
            break
        except Exception as e:
            if "Job was cancelled" in str(e):
                logger.warning(f"Job is cancelled manually.")
                return None
            else:
                logger.warning(f"Job failed because {e}, rerun now.")

    return counts


class QiskitProcessor(object):
    def __init__(
        self,
        use_real_qc=False,
        backend_name=None,
        backend=None,
        noise_model_name=None,
        noise_model=None,
        coupling_map_name=None,
        coupling_map=None,
        basis_gates_name=None,
        basis_gates=None,
        n_shots=8192,
        initial_layout=None,
        seed_transpiler=42,
        seed_simulator=42,
        optimization_level=None,
        max_jobs=5,
        remove_ops=False,
        remove_ops_thres=1e-4,
        transpile_with_ancilla=True,
        hub="ibm-q",
        group="open",
        project="main",
        layout_method=None,
        routing_method=None,
    ):
        self.use_real_qc = use_real_qc
        self.noise_model_name = noise_model_name
        self.backend_name = backend_name
        self.coupling_map_name = coupling_map_name
        self.basis_gates_name = basis_gates_name
        self.n_shots = n_shots
        self.initial_layout = initial_layout
        self.seed_transpiler = seed_transpiler
        self.seed_simulator = seed_simulator
        self.optimization_level = optimization_level
        self.max_jobs = max_jobs
        self.transpile_with_ancilla = transpile_with_ancilla

        self.layout_method = layout_method
        self.routing_method = routing_method

        self.hub = hub
        self.group = group
        self.project = project
        self.backend = backend
        self.provider = None
        self.noise_model = noise_model
        self.coupling_map = coupling_map
        self.basis_gates = basis_gates
        self.properties = None
        self.empty_pass_manager = EmptyPassManager()

        self.transpiled_circs = None

        self.remove_ops = remove_ops
        self.remove_ops_thres = remove_ops_thres

        self.qiskit_init()

    def get_noise_model(self, name):
        if name in IBMQ_NAMES:
            backend = self.provider.get_backend(name)
            self.properties = backend.properties()
            noise_model = NoiseModel.from_backend(backend)
        else:
            noise_model = None

        return noise_model

    def get_coupling_map(self, name):
        if name in IBMQ_NAMES:
            backend = self.provider.get_backend(name)
            coupling_map = backend.configuration().coupling_map
        else:
            if name == "four_all":
                coupling_map = [
                    [0, 1],
                    [1, 0],
                    [0, 2],
                    [2, 0],
                    [0, 3],
                    [3, 0],
                    [1, 2],
                    [2, 1],
                    [1, 3],
                    [3, 1],
                    [2, 3],
                    [3, 2],
                ]
            else:
                coupling_map = None

        return coupling_map

    def get_basis_gates(self, name):
        if name in IBMQ_NAMES:
            backend = self.provider.get_backend(name)
            basis_gates = backend.configuration().basis_gates
        else:
            basis_gates = None

        return basis_gates

    def qiskit_init(self):
        self.provider = None
        self.properties = None

        if self.backend is None:
            # initialize now
            IBMQ.load_account()
            self.provider = get_provider_hub_group_project(
                hub=self.hub,
                group=self.group,
                project=self.project,
            )
            if self.use_real_qc:
                self.backend = self.provider.get_backend(self.backend_name)
                self.properties = self.backend.properties()
                self.coupling_map = self.get_coupling_map(self.backend_name)
            else:
                # use simulator
                self.backend = Aer.get_backend(
                    "qasm_simulator", max_parallel_experiments=0
                )
                self.noise_model = self.get_noise_model(self.noise_model_name)
                self.coupling_map = self.get_coupling_map(self.coupling_map_name)
                self.basis_gates = self.get_basis_gates(self.basis_gates_name)
        else:
            # predefined backend
            self.backend_name = self.backend.name()
            print(f"Use backend: {self.backend_name}")
            if self.coupling_map is None:
                self.coupling_map = self.backend.configuration().coupling_map
            if self.basis_gates is None:
                self.basis_gates = self.backend.configuration().basis_gates

    def set_layout(self, layout):
        self.initial_layout = layout

    def set_backend(self, backend):
        self.backend = backend

    def transpile(self, circs):
        if not self.transpile_with_ancilla and self.coupling_map is not None:
            # only use same number of physical qubits as virtual qubits
            # !! the risk is that the remaining graph is not a connected graph,
            # need fix this later
            coupling_map = []
            for pair in self.coupling_map:
                if all([p_wire < len(circs.qubits) for p_wire in pair]):
                    coupling_map.append(pair)
        else:
            coupling_map = self.coupling_map
        transpiled_circs = transpile(
            circuits=circs,
            backend=self.backend,
            basis_gates=self.basis_gates,
            coupling_map=coupling_map,
            initial_layout=self.initial_layout,
            seed_transpiler=self.seed_transpiler,
            optimization_level=self.optimization_level,
        )
        return transpiled_circs

    def preprocess_parameterized(
        self,
        q_device,
        q_layer_parameterized,
        q_layer_fixed,
        q_layer_measure,
        x,
    ):
        circ_parameterized, params = tq2qiskit_parameterized(
            q_device, q_layer_parameterized.func_list
        )
        circ_fixed = tq2qiskit(
            q_device,
            q_layer_fixed,
            remove_ops=self.remove_ops,
            remove_ops_thres=self.remove_ops_thres,
        )

        circ_measurement = tq2qiskit_measurement(q_device, q_layer_measure)
        circ = circ_parameterized + circ_fixed + circ_measurement

        logger.info(f"Before transpile: {get_circ_stats(circ)}")
        transpiled_circ = self.transpile(circ)
        logger.info(f"After transpile: {get_circ_stats(transpiled_circ)}")
        self.transpiled_circs = [transpiled_circ]
        # construct the parameter_binds
        binds_all = []
        for inputs_single in x:
            binds = {}
            for k, input_single in enumerate(inputs_single):
                binds[params[k]] = input_single.item()
            binds_all.append(binds)

        return transpiled_circ, binds_all

    def process_parameterized(
        self,
        q_device: tq.QuantumDevice,
        q_layer_parameterized: tq.QuantumModule,
        q_layer_fixed: tq.QuantumModule,
        q_layer_measure: tq.QuantumModule,
        x,
        parallel=True,
    ):
        """
        separate the conversion, encoder part will be converted to a
        parameterized Qiskit QuantumCircuit. The remaining part will be a
        non-parameterized QuantumCircuit. In this case, only one time of
        compilation is required.

        q_layer_parameterized needs to have a func_list to specify the gates

        for parallel:
        JobManager has bugs when submitting job, so use multiprocessing instead
        """
        transpiled_circ, binds_all = self.preprocess_parameterized(
            q_device, q_layer_parameterized, q_layer_fixed, q_layer_measure, x
        )

        if parallel:
            if hasattr(self.backend.configuration(), "max_experiments"):
                chunk_size = self.backend.configuration().max_experiments
            else:
                # using simulator, apply multithreading
                chunk_size = len(binds_all) // self.max_jobs

            if chunk_size == 0:
                split_binds = [binds_all]
            else:
                split_binds = [
                    binds_all[i : i + chunk_size]
                    for i in range(0, len(binds_all), chunk_size)
                ]

            qiskit_verbose = self.max_jobs <= 6
            feed_dicts = []
            for split_bind in split_binds:
                feed_dict = {
                    "experiments": transpiled_circ,
                    "backend": self.backend,
                    "pass_manager": self.empty_pass_manager,
                    "shots": self.n_shots,
                    "seed_simulator": self.seed_simulator,
                    "noise_model": self.noise_model,
                    "parameter_binds": split_bind,
                }
                feed_dicts.append([feed_dict, qiskit_verbose])

            p = multiprocessing.Pool(self.max_jobs)
            results = p.map(run_job_worker, feed_dicts)
            p.close()

            if all(isinstance(result, dict) for result in results):
                counts = results
            else:
                if isinstance(results[-1], dict):
                    results[-1] = [results[-1]]
                counts = list(itertools.chain(*results))
        else:
            job = execute(
                experiments=transpiled_circ,
                backend=self.backend,
                pass_manager=self.empty_pass_manager,
                shots=self.n_shots,
                seed_simulator=self.seed_simulator,
                noise_model=self.noise_model,
                parameter_binds=binds_all,
            )
            job_monitor(job, interval=1)

            result = job.result()
            counts = result.get_counts()

        measured_qiskit = get_expectations_from_counts(counts, n_wires=q_device.n_wires)
        measured_qiskit = torch.tensor(measured_qiskit, device=x.device)

        return measured_qiskit

    def preprocess_parameterized_and_shift(
        self,
        q_device,
        q_layer_parameterized,
        q_layer_fixed,
        q_layer_measure,
        x,
        shift_encoder,
        shift_this_step,
    ):
        circ_parameterized, params = tq2qiskit_parameterized(
            q_device, q_layer_parameterized.func_list
        )
        circ_fixed_list = []
        circ_fixed = tq2qiskit(
            q_device,
            q_layer_fixed,
            remove_ops=self.remove_ops,
            remove_ops_thres=self.remove_ops_thres,
        )
        circ_fixed_list.append(circ_fixed)

        # not shift encoder ==> shift fixed layer
        if not shift_encoder:
            for i, named_param in enumerate(q_layer_fixed.named_parameters()):
                if shift_this_step[i]:
                    param = named_param[-1]
                    param.copy_(param + np.pi * 0.5)
                    circ_fixed = tq2qiskit(
                        q_device,
                        q_layer_fixed,
                        remove_ops=self.remove_ops,
                        remove_ops_thres=self.remove_ops_thres,
                    )
                    circ_fixed_list.append(circ_fixed)
                    param.copy_(param - np.pi)
                    circ_fixed = tq2qiskit(
                        q_device,
                        q_layer_fixed,
                        remove_ops=self.remove_ops,
                        remove_ops_thres=self.remove_ops_thres,
                    )
                    circ_fixed_list.append(circ_fixed)
                    param.copy_(param + np.pi * 0.5)

        self.transpiled_circs = []
        for circ_fixed in circ_fixed_list:
            circ = circ_parameterized + circ_fixed
            v_c_reg_mapping = q_layer_measure.v_c_reg_mapping
            if v_c_reg_mapping is not None:
                for q_reg, c_reg in v_c_reg_mapping["v2c"].items():
                    circ.measure(q_reg, c_reg)
            else:
                circ.measure(
                    list(range(q_device.n_wires)), list(range(q_device.n_wires))
                )

            transpiled_circ = self.transpile(circ)
            self.transpiled_circs.append(transpiled_circ)
        # construct the parameter_binds
        binds_all = []
        if shift_encoder:
            for idx in range(x.size()[1]):
                x[:, idx] += np.pi * 0.5
                for inputs_single in x:
                    binds = {}
                    for k, input_single in enumerate(inputs_single):
                        binds[params[k]] = input_single.item()
                    binds_all.append(binds)

                x[:, idx] -= np.pi
                for inputs_single in x:
                    binds = {}
                    for k, input_single in enumerate(inputs_single):
                        binds[params[k]] = input_single.item()
                    binds_all.append(binds)

                x[:, idx] += np.pi * 0.5
        else:
            for inputs_single in x:
                binds = {}
                for k, input_single in enumerate(inputs_single):
                    binds[params[k]] = input_single.item()
                binds_all.append(binds)

        return self.transpiled_circs, binds_all

    def process_parameterized_and_shift(
        self,
        q_device: tq.QuantumDevice,
        q_layer_parameterized: tq.QuantumModule,
        q_layer_fixed: tq.QuantumModule,
        q_layer_measure: tq.QuantumModule,
        x,
        shift_encoder=False,
        parallel=True,
        shift_this_step=None,
    ):
        """
        separate the conversion, encoder part will be converted to a
        parameterized Qiskit QuantumCircuit. The remaining part will be a
        non-parameterized QuantumCircuit. In this case, only one time of
        compilation is required.

        q_layer_parameterized needs to have a func_list to specify the gates

        for parallel:
        JobManager has bugs when submitting job, so use multiprocessing instead
        """
        transpiled_circs, binds_all = self.preprocess_parameterized_and_shift(
            q_device,
            q_layer_parameterized,
            q_layer_fixed,
            q_layer_measure,
            x,
            shift_encoder,
            shift_this_step,
        )

        time_spent_list = []

        if parallel:
            if hasattr(self.backend.configuration(), "max_experiments"):
                chunk_size = self.backend.configuration().max_experiments
            else:
                # using simulator, apply multithreading
                chunk_size = len(binds_all) // self.max_jobs

            split_binds = [
                binds_all[i : i + chunk_size]
                for i in range(0, len(binds_all), chunk_size)
            ]

            qiskit_verbose = self.max_jobs <= 6
            feed_dicts = []
            for split_bind in split_binds:
                feed_dict = {
                    "experiments": transpiled_circs,
                    "backend": self.backend,
                    "pass_manager": self.empty_pass_manager,
                    "shots": self.n_shots,
                    "seed_simulator": self.seed_simulator,
                    "noise_model": self.noise_model,
                    "parameter_binds": split_bind,
                }
                feed_dicts.append([feed_dict, qiskit_verbose])

            p = multiprocessing.Pool(self.max_jobs)
            results = p.map(run_job_worker, feed_dicts)
            p.close()

            if all(isinstance(result, dict) for result in results):
                counts = results
            else:
                if isinstance(results[-1], dict):
                    results[-1] = [results[-1]]
                counts = list(itertools.chain(*results))
        else:
            chunk_size = 75 // len(binds_all)
            split_circs = [
                transpiled_circs[i : i + chunk_size]
                for i in range(0, len(transpiled_circs), chunk_size)
            ]
            counts = []
            total_time_spent = datetime.timedelta()
            total_cont = 0
            for circ in split_circs:
                while True:
                    try:
                        job = execute(
                            experiments=circ,
                            backend=self.backend,
                            pass_manager=self.empty_pass_manager,
                            shots=self.n_shots,
                            seed_simulator=self.seed_simulator,
                            noise_model=self.noise_model,
                            parameter_binds=binds_all,
                        )
                        job_monitor(job, interval=1)
                        result = (
                            job.result()
                        )  # qiskit.providers.ibmq.job.exceptions.IBMQJobFailureError:Job has failed. Use the error_message() method to get more details
                        counts = counts + result.get_counts()
                        # time_per_step = job.time_per_step()
                        # time_spent = time_per_step['COMPLETED'] - time_per_step['RUNNING'] + time_per_step['QUEUED'] - job.time_per_step()['CREATING']
                        # time_spent_list.append(time_spent)
                        # print(time_spent)
                        # total_time_spent += time_spent
                        # total_cont += 1
                        # print(total_time_spent / total_cont)
                        break
                    except (QiskitError) as e:
                        logger.warning("Job failed, rerun now.")
                        print(e.message)

        measured_qiskit = get_expectations_from_counts(counts, n_wires=q_device.n_wires)
        measured_qiskit = torch.tensor(measured_qiskit, device=x.device)

        return measured_qiskit, time_spent_list

    def process_multi_measure(
        self,
        q_device: tq.QuantumDevice,
        q_layer: tq.QuantumModule,
        q_layer_measure: tq.QuantumModule,
    ):
        obs_list = q_layer_measure.obs_list
        circ_fixed = tq2qiskit(
            q_device,
            q_layer,
            remove_ops=self.remove_ops,
            remove_ops_thres=self.remove_ops_thres,
        )

        transpiled_circ_fixed = self.transpile(circ_fixed)

        circ_all = []

        for hamil in obs_list:
            circ_diagonalize = QuantumCircuit(q_device.n_wires, q_device.n_wires)

            # diagonalize the measurements
            for wire, observable in zip(hamil["wires"], hamil["observables"]):
                if observable == "x":
                    circ_diagonalize.h(qubit=wire)
                elif observable == "y":
                    circ_diagonalize.z(qubit=wire)
                    circ_diagonalize.s(qubit=wire)
                    circ_diagonalize.h(qubit=wire)

            circ_measurement = tq2qiskit_measurement(q_device, q_layer_measure)

            circ_diagonalize = circ_diagonalize + circ_measurement

            transpiled_circ_diagonalize = self.transpile(circ_diagonalize)
            circ_all.append(transpiled_circ_fixed + transpiled_circ_diagonalize)

        self.transpiled_circs = circ_all

        if hasattr(self.backend.configuration(), "max_experiments"):
            chunk_size = self.backend.configuration().max_experiments
        else:
            # using simulator, apply multithreading
            chunk_size = len(circ_all) // self.max_jobs

        split_circs = [
            circ_all[i : i + chunk_size] for i in range(0, len(circ_all), chunk_size)
        ]

        qiskit_verbose = self.max_jobs <= 2
        feed_dicts = []
        for split_circ in split_circs:
            feed_dict = {
                "experiments": split_circ,
                "backend": self.backend,
                "pass_manager": self.empty_pass_manager,
                "shots": self.n_shots,
                "seed_simulator": self.seed_simulator,
                "noise_model": self.noise_model,
            }
            feed_dicts.append([feed_dict, qiskit_verbose])

        p = multiprocessing.Pool(self.max_jobs)
        results = p.map(run_job_worker, feed_dicts)
        p.close()

        if all(isinstance(result, dict) for result in results):
            counts = results
        else:
            if isinstance(results[-1], dict):
                results[-1] = [results[-1]]
            counts = list(itertools.chain(*results))

        measured_qiskit = get_expectations_from_counts(counts, n_wires=q_device.n_wires)

        measured_qiskit = torch.tensor(measured_qiskit, device=q_device.state.device)

        return measured_qiskit

    def process(
        self,
        q_device: tq.QuantumDevice,
        q_layer: tq.QuantumModule,
        q_layer_measure: tq.QuantumModule,
        x,
    ):
        circs = []
        for i, x_single in tqdm(enumerate(x)):
            circ = tq2qiskit(q_device, q_layer, x_single.unsqueeze(0))
            circ_measurement = tq2qiskit_measurement(q_device, q_layer_measure)
            circ = circ + circ_measurement

            circs.append(circ)

        transpiled_circs = self.transpile(circs)
        self.transpiled_circs = transpiled_circs

        job = execute(
            experiments=transpiled_circs,
            backend=self.backend,
            shots=self.n_shots,
            # initial_layout=self.initial_layout,
            seed_transpiler=self.seed_transpiler,
            seed_simulator=self.seed_simulator,
            coupling_map=self.coupling_map,
            basis_gates=self.basis_gates,
            noise_model=self.noise_model,
            optimization_level=self.optimization_level,
        )
        job_monitor(job, interval=1)

        result = job.result()
        counts = result.get_counts()

        measured_qiskit = get_expectations_from_counts(counts, n_wires=q_device.n_wires)
        measured_qiskit = torch.tensor(measured_qiskit, device=x.device)

        return measured_qiskit

    def process_ready_circs(self, q_device, circs_all, parallel=True):
        circs_all_transpiled = []
        for circ in tqdm(circs_all):
            circs_all_transpiled.append(self.transpile(circ))

        circs_all = circs_all_transpiled

        if parallel:
            if hasattr(self.backend.configuration(), "max_experiments"):
                chunk_size = self.backend.configuration().max_experiments
            else:
                # using simulator, apply multithreading
                chunk_size = len(circs_all) // self.max_jobs
                if chunk_size == 0:
                    chunk_size = 1

            split_circs = [
                circs_all[i : i + chunk_size]
                for i in range(0, len(circs_all), chunk_size)
            ]

            qiskit_verbose = self.max_jobs <= 6
            feed_dicts = []
            for split_circ in split_circs:
                feed_dict = {
                    "experiments": split_circ,
                    "backend": self.backend,
                    "pass_manager": self.empty_pass_manager,
                    "shots": self.n_shots,
                    "seed_simulator": self.seed_simulator,
                    "noise_model": self.noise_model,
                }
                feed_dicts.append([feed_dict, qiskit_verbose])

            p = multiprocessing.Pool(self.max_jobs)
            results = p.map(run_job_worker, feed_dicts)
            p.close()

            if all(isinstance(result, dict) for result in results):
                counts = results
            else:
                if isinstance(results[-1], dict):
                    results[-1] = [results[-1]]
                counts = list(itertools.chain(*results))
        else:
            job = execute(
                experiments=circs_all,
                backend=self.backend,
                pass_manager=self.empty_pass_manager,
                shots=self.n_shots,
                seed_simulator=self.seed_simulator,
                noise_model=self.noise_model,
            )
            job_monitor(job, interval=1)

            result = job.result()
            counts = result.get_counts()

        measured_qiskit = get_expectations_from_counts(counts, n_wires=q_device.n_wires)
        measured_torch = torch.tensor(measured_qiskit)

        return measured_torch



import torch
import torchquantum as tq
import numpy as np

from typing import Any


class QuantizeFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, x: torch.Tensor) -> Any:
        # should be round so that the changes would be small, values close to
        # 2pi should go to 2pi
        return x.round()

    @staticmethod
    def backward(ctx: Any, grad_output: Any) -> Any:
        grad_input = grad_output.clone()
        mean, std = grad_input.mean(), grad_input.std()
        return grad_input.clamp_(mean - 3 * std, mean + 3 * std)


class CliffordQuantizer(object):
    def __init__(self):
        pass

    # straight-through estimator
    @staticmethod
    def quantize_sse(params):
        param = params[0][0]
        param = param % (2 * np.pi)
        param = np.pi / 2 * QuantizeFunction.apply(param / (np.pi / 2))
        params = param.unsqueeze(0).unsqueeze(0)
        return params



import torch
import torch.nn as nn
import numpy as np

from torchquantum.macro import C_DTYPE
from torchquantum.functional import func_name_dict

from typing import Union

__all__ = ["QuantumDevice"]


class QuantumDevice(nn.Module):
    def __init__(
        self,
        n_wires: int,
        device_name: str = "default",
        bsz: int = 1,
        device: Union[torch.device, str] = "cpu",
        record_op: bool = False,
    ):
        """A quantum device that contains the quantum state vector.
        Args:
            n_wires: number of qubits
            device_name: name of the quantum device
            bsz: batch size of the quantum state
            device: which classical computing device to use, 'cpu' or 'cuda'
            record_op: whether to record the operations on the quantum device and then
                they can be used to construct a static computation graph
        """
        super().__init__()
        # number of qubits
        # the states are represented in a multi-dimension tensor
        # from left to right: qubit 0 to n
        self.n_wires = n_wires
        self.device_name = device_name
        self.bsz = bsz
        self.device = device

        _state = torch.zeros(2**self.n_wires, dtype=C_DTYPE)
        _state[0] = 1 + 0j  # type: ignore
        _state = torch.reshape(_state, [2] * self.n_wires).to(self.device)
        self.register_buffer("state", _state)

        repeat_times = [bsz] + [1] * len(self.state.shape)  # type: ignore
        self._states = self.state.repeat(*repeat_times)  # type: ignore
        self.register_buffer("states", self._states)

        self.record_op = record_op
        self.op_history = []

    def reset_op_history(self):
        self.op_history = []

    def clone_states(self, existing_states: torch.Tensor):
        """Clone the states of the quantum device."""
        self.states = existing_states.clone()

    def set_states(self, states: torch.Tensor):
        """Set the states of the quantum device. The states are represented"""
        bsz = states.shape[0]
        self.states = torch.reshape(states, [bsz] + [2] * self.n_wires)

    def reset_states(self, bsz: int):
        repeat_times = [bsz] + [1] * len(self.state.shape)
        self.states = self.state.repeat(*repeat_times).to(self.state.device)

    def reset_identity_states(self):
        """Make the states as the identity matrix, one dim is the batch
        dim. Useful for verification.
        """
        self.states = torch.eye(
            2**self.n_wires, device=self.state.device, dtype=C_DTYPE
        ).reshape([2**self.n_wires] + [2] * self.n_wires)

    def reset_all_eq_states(self, bsz: int):
        """Make the states as the equal superposition state, one dim is the
        batch dim. Useful for verification.
        """
        energy = np.sqrt(1 / (2**self.n_wires) / 2)
        all_eq_state = torch.ones(2**self.n_wires, dtype=C_DTYPE) * (
            energy + energy * 1j
        )
        all_eq_state = all_eq_state.reshape([2] * self.n_wires)
        repeat_times = [bsz] + [1] * len(self.state.shape)
        self.states = all_eq_state.repeat(*repeat_times).to(self.state.device)

    def get_states_1d(self):
        """Return the states in a 1d tensor."""
        bsz = self.states.shape[0]
        return torch.reshape(self.states, [bsz, 2**self.n_wires])

    def get_state_1d(self):
        """Return the state in a 1d tensor."""
        return torch.reshape(self.state, [2**self.n_wires])

    @property
    def name(self):
        """Return the name of the device."""
        return self.__class__.__name__

    def __repr__(self):
        return f" class: {self.name} \n device name: {self.device_name} \n number of qubits: {self.n_wires} \n batch size: {self.bsz} \n current computing device: {self.state.device} \n recording op history: {self.record_op} \n current states: {self.get_states_1d().cpu().detach().numpy()}"


for func_name, func in func_name_dict.items():
    setattr(QuantumDevice, func_name, func)



import argparse
import pdb
import torch
import torchquantum as tq
import numpy as np

from qiskit import Aer, execute
from torchpack.utils.logging import logger
from torchquantum.utils import (
    switch_little_big_endian_matrix,
    switch_little_big_endian_state,
    get_expectations_from_counts,
    find_global_phase,
)
from test.static_mode_test import QLayer as AllRandomLayer
from torchquantum.plugins import tq2qiskit
from torchquantum.macro import F_DTYPE


def unitary_tq_vs_qiskit_test():
    for n_wires in range(2, 10):
        q_dev = tq.QuantumDevice(n_wires=n_wires)
        x = torch.randn((1, 100000), dtype=F_DTYPE)
        q_layer = AllRandomLayer(
            n_wires=n_wires,
            wires=list(range(n_wires)),
            n_ops_rd=500,
            n_ops_cin=500,
            n_funcs=500,
            qiskit_compatible=True,
        )

        unitary_tq = q_layer.get_unitary(q_dev, x)
        unitary_tq = switch_little_big_endian_matrix(unitary_tq.data.numpy())

        # qiskit
        circ = tq2qiskit(q_layer, x)
        simulator = Aer.get_backend("unitary_simulator")
        result = execute(circ, simulator).result()
        unitary_qiskit = result.get_unitary(circ)

        stable_threshold = 1e-5
        try:
            # WARNING: need to remove the global phase! The qiskit simulated
            # results sometimes has global phase shift.
            global_phase = find_global_phase(
                unitary_tq, unitary_qiskit, stable_threshold
            )

            if global_phase is None:
                logger.exception(
                    f"Cannot find a stable enough factor to "
                    f"reduce the global phase, increase the "
                    f"stable_threshold and try again"
                )
                raise RuntimeError

            assert np.allclose(unitary_tq * global_phase, unitary_qiskit, atol=1e-6)
            logger.info(f"PASS tq vs qiskit [n_wires]={n_wires}")

        except AssertionError:
            logger.exception(f"FAIL tq vs qiskit [n_wires]={n_wires}")
            raise AssertionError

        except RuntimeError:
            raise RuntimeError

    logger.info(f"PASS tq vs qiskit unitary test")


def state_tq_vs_qiskit_test():
    bsz = 1
    for n_wires in range(2, 10):
        q_dev = tq.QuantumDevice(n_wires=n_wires)
        q_dev.reset_states(bsz=bsz)

        x = torch.randn((1, 100000), dtype=F_DTYPE)
        q_layer = AllRandomLayer(
            n_wires=n_wires,
            wires=list(range(n_wires)),
            n_ops_rd=500,
            n_ops_cin=500,
            n_funcs=500,
            qiskit_compatible=True,
        )

        q_layer(q_dev, x)
        state_tq = q_dev.states.reshape(bsz, -1)
        state_tq = switch_little_big_endian_state(state_tq.data.numpy())

        # qiskit
        circ = tq2qiskit(q_layer, x)
        # Select the StatevectorSimulator from the Aer provider
        simulator = Aer.get_backend("statevector_simulator")

        # Execute and get counts
        result = execute(circ, simulator).result()
        state_qiskit = result.get_statevector(circ)

        stable_threshold = 1e-5
        try:
            # WARNING: need to remove the global phase! The qiskit simulated
            # results sometimes has global phase shift.
            global_phase = find_global_phase(
                state_tq, np.expand_dims(state_qiskit, 0), stable_threshold
            )

            if global_phase is None:
                logger.exception(
                    f"Cannot find a stable enough factor to "
                    f"reduce the global phase, increase the "
                    f"stable_threshold and try again"
                )
                raise RuntimeError

            assert np.allclose(state_tq * global_phase, state_qiskit, atol=1e-6)
            logger.info(f"PASS tq vs qiskit [n_wires]={n_wires}")

        except AssertionError:
            logger.exception(f"FAIL tq vs qiskit [n_wires]={n_wires}")
            raise AssertionError

        except RuntimeError:
            raise RuntimeError

    logger.info(f"PASS tq vs qiskit statevector test")


def measurement_tq_vs_qiskit_test():
    bsz = 1
    for n_wires in range(2, 10):
        q_dev = tq.QuantumDevice(n_wires=n_wires)
        q_dev.reset_states(bsz=bsz)

        x = torch.randn((1, 100000), dtype=F_DTYPE)
        q_layer = AllRandomLayer(
            n_wires=n_wires,
            wires=list(range(n_wires)),
            n_ops_rd=500,
            n_ops_cin=500,
            n_funcs=500,
            qiskit_compatible=True,
        )

        q_layer(q_dev, x)
        measurer = tq.MeasureAll(obs=tq.PauliZ)
        # flip because qiskit is from N to 0, tq is from 0 to N
        measured_tq = np.flip(measurer(q_dev).data[0].numpy())

        # qiskit
        circ = tq2qiskit(q_layer, x)
        circ.measure(list(range(n_wires)), list(range(n_wires)))

        # Select the QasmSimulator from the Aer provider
        simulator = Aer.get_backend("qasm_simulator")

        # Execute and get counts
        result = execute(circ, simulator, shots=1000000).result()
        counts = result.get_counts(circ)
        measured_qiskit = get_expectations_from_counts(counts, n_wires=n_wires)

        try:
            # WARNING: the measurement has randomness, so tolerate larger
            # differences (MAX 20%) between tq and qiskit
            # typical mean difference is less than 1%
            diff = np.abs(measured_tq - measured_qiskit).mean()
            diff_ratio = (
                np.abs((measured_tq - measured_qiskit) / measured_qiskit)
            ).mean()
            logger.info(f"Diff: tq vs qiskit {diff} \t Diff Ratio: " f"{diff_ratio}")
            assert np.allclose(measured_tq, measured_qiskit, atol=1e-4, rtol=2e-1)
            logger.info(f"PASS tq vs qiskit [n_wires]={n_wires}")

        except AssertionError:
            logger.exception(f"FAIL tq vs qiskit [n_wires]={n_wires}")
            raise AssertionError

    logger.info(f"PASS tq vs qiskit measurement test")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--pdb", action="store_true", help="pdb")
    args = parser.parse_args()

    # seed = 45
    # torch.manual_seed(seed)
    # np.random.seed(seed)

    if args.pdb:
        pdb.set_trace()

    unitary_tq_vs_qiskit_test()
    state_tq_vs_qiskit_test()
    measurement_tq_vs_qiskit_test()



import torch.nn as nn
import torch.nn.utils.prune

from typing import Any, Callable, Dict
from torchpack.train import Trainer
from torchpack.utils.typing import Optimizer, Scheduler
from torchpack.utils.config import configs
from torchquantum.utils import get_unitary_loss, build_module_op_list
from torchpack.utils.logging import logger
from torchpack.callbacks.writers import TFEventWriter
from torchpack.train.exception import StopTraining

__all__ = ["QTrainer", "ParamsShiftTrainer"]


class QTrainer(Trainer):
    def __init__(
        self,
        *,
        model: nn.Module,
        criterion: Callable,
        optimizer: Optimizer,
        scheduler: Scheduler,
    ) -> None:
        self.model = model
        self.legalized_model = None
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.solution = None
        self.score = None

    def _before_epoch(self) -> None:
        self.model.train()

    def run_step(self, feed_dict: Dict[str, Any], legalize=False) -> Dict[str, Any]:
        output_dict = self._run_step(feed_dict, legalize=legalize)
        return output_dict

    def _run_step(self, feed_dict: Dict[str, Any], legalize=False) -> Dict[str, Any]:
        if configs.run.device == "gpu":
            inputs = feed_dict[configs.dataset.input_name].cuda(non_blocking=True)
            targets = feed_dict[configs.dataset.target_name].cuda(non_blocking=True)
        else:
            inputs = feed_dict[configs.dataset.input_name]
            targets = feed_dict[configs.dataset.target_name]
        if legalize:
            outputs = self.legalized_model(inputs)
        else:
            outputs = self.model(inputs)
        loss = self.criterion(outputs, targets)
        nll_loss = loss.item()
        unitary_loss = 0

        if configs.regularization.unitary_loss:
            unitary_loss = get_unitary_loss(self.model)
            if configs.regularization.unitary_loss_lambda_trainable:
                loss += self.model.unitary_loss_lambda[0] * unitary_loss
            else:
                loss += configs.regularization.unitary_loss_lambda * unitary_loss

        if loss.requires_grad:
            for k, group in enumerate(self.optimizer.param_groups):
                self.summary.add_scalar(f"lr/lr_group{k}", group["lr"])
            self.summary.add_scalar("loss", loss.item())
            self.summary.add_scalar("nll_loss", nll_loss)
            if getattr(self.model, "sample_arch", None) is not None:
                for writer in self.summary.writers:
                    if isinstance(writer, TFEventWriter):
                        writer.writer.add_text(
                            "sample_arch", str(self.model.sample_arch), self.global_step
                        )

            if configs.regularization.unitary_loss:
                if configs.regularization.unitary_loss_lambda_trainable:
                    self.summary.add_scalar(
                        "u_loss_lambda", self.model.unitary_loss_lambda.item()
                    )
                else:
                    self.summary.add_scalar(
                        "u_loss_lambda", configs.regularization.unitary_loss_lambda
                    )
                self.summary.add_scalar("u_loss", unitary_loss.item())

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        return {"outputs": outputs, "targets": targets}

    def _after_epoch(self) -> None:
        self.model.eval()
        self.scheduler.step()
        # if configs.legalization.legalize:
        #     if self.epoch_num % configs.legalization.epoch_interval == 0:
        #         legalize_unitary(self.model)

    def _after_step(self, output_dict) -> None:
        pass
        # if configs.legalization.legalize:
        #     if self.global_step % configs.legalization.step_interval == 0:
        #         legalize_unitary(self.model)

    def _state_dict(self) -> Dict[str, Any]:
        state_dict = dict()
        # need to store model arch because of randomness of random layers
        state_dict["model_arch"] = self.model
        state_dict["model"] = self.model.state_dict()
        state_dict["optimizer"] = self.optimizer.state_dict()
        state_dict["scheduler"] = self.scheduler.state_dict()
        if getattr(self.model, "sample_arch", None) is not None:
            state_dict["sample_arch"] = self.model.sample_arch

        if self.solution is not None:
            state_dict["solution"] = self.solution
            state_dict["score"] = self.score

        if getattr(self.model, "encoder", None) is not None:
            if getattr(self.model.encoder, "func_list", None) is not None:
                state_dict["encoder_func_list"] = self.model.encoder.func_list

        if getattr(self.model, "q_layer", None) is not None:
            state_dict["q_layer_op_list"] = build_module_op_list(self.model.q_layer)

        if getattr(self.model, "measure", None) is not None:
            if getattr(self.model.measure, "v_c_reg_mapping", None) is not None:
                state_dict["v_c_reg_mapping"] = self.model.measure.v_c_reg_mapping

        if getattr(self.model, "nodes", None) is not None:
            state_dict["encoder_func_list"] = [
                node.encoder.func_list for node in self.model.nodes
            ]
            if configs.model.transpile_before_run:
                # only save op_list if transpile before run
                state_dict["q_layer_op_list"] = [
                    build_module_op_list(node.q_layer) for node in self.model.nodes
                ]
            state_dict["v_c_reg_mapping"] = [
                node.measure.v_c_reg_mapping for node in self.model.nodes
            ]
        for attr in ["v_c_reg_mapping", "encoder_func_list", "q_layer_op_list"]:
            if state_dict.get(attr, None) is None:
                logger.warning(f"No {attr} found, will not save it.")

        return state_dict

    def _load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        # self.model.load_state_dict(state_dict['model'])
        self.optimizer.load_state_dict(state_dict["optimizer"])
        self.scheduler.load_state_dict(state_dict["scheduler"])


class ParamsShiftTrainer(Trainer):
    def __init__(
        self,
        *,
        model: nn.Module,
        criterion: Callable,
        optimizer: Optimizer,
        scheduler: Scheduler,
    ) -> None:
        self.model = model
        self.legalized_model = None
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.solution = None
        self.score = None
        self.is_training = False
        self.save_global_step = 0
        self.max_acc = 0
        self.max_acc_epoch = 0
        self.global_epoch = 0
        self.work_from_step = 0
        self.grad_dict = None
        self.processor_list = []
        self.profile_shots = False
        self.per_steps = 100
        self.record = dict()

    def load_grad(self, stop_step, grad_dict):
        self.work_from_step = stop_step
        self.grad_dict = grad_dict

    def set_processor_list(self, processor_list):
        self.processor_list = processor_list
        self.profile_shots = True
        self.per_steps = 25

    def set_use_qiskit(self, configs):
        self.use_qiskit_train = configs.qiskit.use_qiskit_train
        self.use_qiskit_valid = configs.qiskit.use_qiskit_valid

    def _before_epoch(self) -> None:
        self.model.train()
        self.is_training = True

    def run_step(self, feed_dict: Dict[str, Any], legalize=False) -> Dict[str, Any]:
        if (
            self.profile_shots
            and self.global_step % self.per_steps == 0
            and self.is_training
        ):
            self.record[self.global_step] = {}
            for i in range(4):
                for shots, processor in self.processor_list:
                    self.model.set_qiskit_processor(processor)
                    self.record[self.global_step][i] = self._get_grad(
                        feed_dict, use_qiskit=True
                    )
            output_dict = self._run_params_shift_step(feed_dict, use_qiskit=False)
            grad_list = []
            for param in self.model.parameters():
                grad_list.append(param.grad.item())
            self.record[self.global_step]["classical"] = grad_list

            return output_dict

        if self.is_training:
            output_dict = self._run_params_shift_step(
                feed_dict, use_qiskit=self.use_qiskit_train
            )
        else:
            output_dict = self._run_step(
                feed_dict, legalize=legalize, use_qiskit=self.use_qiskit_valid
            )
        return output_dict

    def _run_params_shift_step(
        self, feed_dict: Dict[str, Any], use_qiskit=False
    ) -> Dict[str, Any]:
        if configs.run.device == "gpu":
            inputs = feed_dict[configs.dataset.input_name].cuda(non_blocking=True)
            targets = feed_dict[configs.dataset.target_name].cuda(non_blocking=True)
        else:
            inputs = feed_dict[configs.dataset.input_name]
            targets = feed_dict[configs.dataset.target_name]

        # print('bp:')
        # outputs = self.model(inputs)
        # loss = self.criterion(outputs, targets)
        # self.optimizer.zero_grad()
        # loss.backward()
        # for param in self.model.parameters():
        #     print(param.grad)

        outputs = None
        if self.global_step > self.work_from_step:
            outputs = self.model.shift_and_run(
                inputs,
                self.global_step,
                self.steps_per_epoch * self.num_epochs,
                verbose=False,
                use_qiskit=use_qiskit,
            )
        else:
            logger.info(
                "global_step {0}, skip parameters shift.".format(self.global_step)
            )
            outputs = self.model(inputs, False, use_qiskit=use_qiskit)
        loss = self.criterion(outputs, targets)
        nll_loss = loss.item()

        for k, group in enumerate(self.optimizer.param_groups):
            self.summary.add_scalar(f"lr/lr_group{k}", group["lr"])
        self.summary.add_scalar("loss", loss.item())
        self.summary.add_scalar("nll_loss", nll_loss)

        if self.global_step > self.work_from_step:
            self.optimizer.zero_grad()
            loss.backward()
            self.model.backprop_grad()
        else:
            logger.info(
                "global_step {0}, load grad from tensorboard file!".format(
                    self.global_step
                )
            )
            for i, param in enumerate(self.model.parameters()):
                param.grad = (
                    torch.tensor(self.grad_dict[self.global_step][i])
                    .to(dtype=torch.float32, device=param.device)
                    .view(param.shape)
                )
        # print('ps:')
        # for param in self.model.parameters():
        #     print(param.grad)
        self.optimizer.step()

        return {"outputs": outputs, "targets": targets}

    def _get_grad(self, feed_dict: Dict[str, Any], use_qiskit=False) -> Dict[str, Any]:
        if configs.run.device == "gpu":
            inputs = feed_dict[configs.dataset.input_name].cuda(non_blocking=True)
            targets = feed_dict[configs.dataset.target_name].cuda(non_blocking=True)
        else:
            inputs = feed_dict[configs.dataset.input_name]
            targets = feed_dict[configs.dataset.target_name]

        # outputs = self.model(inputs)
        # loss = self.criterion(outputs, targets)
        # self.optimizer.zero_grad()
        # loss.backward()
        # for param in self.model.parameters():
        #     print(param.grad)

        outputs = None
        if self.global_step > self.work_from_step:
            outputs = self.model.shift_and_run(
                inputs,
                self.global_step,
                self.steps_per_epoch * self.num_epochs,
                verbose=False,
                use_qiskit=use_qiskit,
            )
        else:
            logger.info(
                "global_step {0}, skip parameters shift.".format(self.global_step)
            )
            outputs = self.model(inputs, False, use_qiskit)
        loss = self.criterion(outputs, targets)
        nll_loss = loss.item()

        for k, group in enumerate(self.optimizer.param_groups):
            self.summary.add_scalar(f"lr/lr_group{k}", group["lr"])
        self.summary.add_scalar("loss", loss.item())
        self.summary.add_scalar("nll_loss", nll_loss)

        self.optimizer.zero_grad()
        loss.backward()
        if self.global_step > self.work_from_step:
            self.model.backprop_grad()
        else:
            logger.info(
                "global_step {0}, load grad from tensorboard file!".format(
                    self.global_step
                )
            )
            for i, param in enumerate(self.model.parameters()):
                param.grad = (
                    torch.tensor(self.grad_dict[self.global_step][i])
                    .to(dtype=torch.float32, device=param.device)
                    .view(param.shape)
                )
        grad_list = []
        for param in self.model.parameters():
            grad_list.append(param.grad.item())
        # self.optimizer.step()
        # with torch.no_grad():
        #     for param in self.model.parameters():
        #         param.copy_(param - param.grad * 1.0)

        return grad_list

    def _run_step(
        self, feed_dict: Dict[str, Any], legalize=False, use_qiskit=False
    ) -> Dict[str, Any]:
        if configs.run.device == "gpu":
            inputs = feed_dict[configs.dataset.input_name].cuda(non_blocking=True)
            targets = feed_dict[configs.dataset.target_name].cuda(non_blocking=True)
        else:
            inputs = feed_dict[configs.dataset.input_name]
            targets = feed_dict[configs.dataset.target_name]
        if legalize:
            outputs = self.legalized_model(inputs)
        else:
            outputs = self.model(inputs, legalize, use_qiskit)
        loss = self.criterion(outputs, targets)
        nll_loss = loss.item()

        if loss.requires_grad:
            for k, group in enumerate(self.optimizer.param_groups):
                self.summary.add_scalar(f"lr/lr_group{k}", group["lr"])
            self.summary.add_scalar("loss", loss.item())
            self.summary.add_scalar("nll_loss", nll_loss)
            if getattr(self.model, "sample_arch", None) is not None:
                for writer in self.summary.writers:
                    if isinstance(writer, TFEventWriter):
                        writer.writer.add_text(
                            "sample_arch", str(self.model.sample_arch), self.global_step
                        )

            if configs.regularization.unitary_loss:
                if configs.regularization.unitary_loss_lambda_trainable:
                    self.summary.add_scalar(
                        "u_loss_lambda", self.model.unitary_loss_lambda.item()
                    )
                else:
                    self.summary.add_scalar(
                        "u_loss_lambda", configs.regularization.unitary_loss_lambda
                    )
                self.summary.add_scalar("u_loss", unitary_loss.item())

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        return {"outputs": outputs, "targets": targets}

    def _after_epoch(self) -> None:
        self.model.eval()
        self.is_training = False
        self.scheduler.step()
        # if configs.legalization.legalize:
        #     if self.epoch_num % configs.legalization.epoch_interval == 0:
        #         legalize_unitary(self.model)
        self.save_global_step = self.global_step
        self.global_step = int(self.model.num_forwards)

    def _trigger_epoch(self) -> None:
        self.global_step = self.save_global_step

    def _after_step(self, output_dict) -> None:
        pass
        # if configs.legalization.legalize:
        #     if self.global_step % configs.legalization.step_interval == 0:
        #         legalize_unitary(self.model)

    def _state_dict(self) -> Dict[str, Any]:
        state_dict = dict()
        # need to store model arch because of randomness of random layers
        # state_dict['model_arch'] = self.model
        state_dict["model"] = self.model.state_dict()
        state_dict["optimizer"] = self.optimizer.state_dict()
        state_dict["scheduler"] = self.scheduler.state_dict()
        if getattr(self.model, "sample_arch", None) is not None:
            state_dict["sample_arch"] = self.model.sample_arch

        if self.solution is not None:
            state_dict["solution"] = self.solution
            state_dict["score"] = self.score

        if getattr(self.model, "encoder", None) is not None:
            if getattr(self.model.encoder, "func_list", None) is not None:
                state_dict["encoder_func_list"] = self.model.encoder.func_list

        if getattr(self.model, "q_layer", None) is not None:
            state_dict["q_layer_op_list"] = build_module_op_list(self.model.q_layer)

        if getattr(self.model, "measure", None) is not None:
            if getattr(self.model.measure, "v_c_reg_mapping", None) is not None:
                state_dict["v_c_reg_mapping"] = self.model.measure.v_c_reg_mapping

        if getattr(self.model, "nodes", None) is not None:
            state_dict["encoder_func_list"] = [
                node.encoder.func_list for node in self.model.nodes
            ]
            if configs.model.transpile_before_run:
                # only save op_list if transpile before run
                state_dict["q_layer_op_list"] = [
                    build_module_op_list(node.q_layer) for node in self.model.nodes
                ]
            state_dict["v_c_reg_mapping"] = [
                node.measure.v_c_reg_mapping for node in self.model.nodes
            ]
        for attr in ["v_c_reg_mapping", "encoder_func_list", "q_layer_op_list"]:
            if state_dict.get(attr, None) is None:
                logger.warning(f"No {attr} found, will not save it.")

        return state_dict

    def _load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        # self.model.load_state_dict(state_dict['model'])
        self.optimizer.load_state_dict(state_dict["optimizer"])
        self.scheduler.load_state_dict(state_dict["scheduler"])

    def update_accuracy(self, name, accuracy):
        if name == "acc/test":
            self.global_epoch += 1
            if accuracy > self.max_acc:
                self.max_acc = accuracy
                self.max_acc_epoch = self.global_epoch
            elif self.global_epoch - self.max_acc_epoch >= 50:
                raise StopTraining(
                    "Early stop at epoch{0}, num inferences = {1}, global_step={2}".format(
                        self.global_epoch, self.global_step, self.save_global_step
                    )
                )



from setuptools import setup, find_packages

VERSION = {}  # type: ignore

with open("torchquantum/__version__.py", "r") as version_file:
    exec(version_file.read(), VERSION)

if __name__ == "__main__":
    setup(
        name="torchquantum",
        version=VERSION["version"],
        description="A PyTorch-based framework for differentiable classical simulation of quantum computing",
        url="https://github.com/mit-han-lab/torchquantum",
        author="Hanrui Wang, Jiannan Cao, Jessica Ding, Jiai Gu, Song Han, Zirui Li, Zhiding Liang, Pengyu Liu, Mohammadreza Tavasoli",
        author_email="hanruiwang.hw@gmail.com",
        license="MIT",
        install_requires=[
            "numpy>=1.19.2",
            "torchvision>=0.9.0.dev20210130",
            "tqdm>=4.56.0",
            "setuptools>=52.0.0",
            "torch>=1.8.0",
            "torchpack>=0.3.0",
            "qiskit==0.38.0",
            "matplotlib>=3.3.2",
            "pathos>=0.2.7",
            "pylatexenc>=2.10",
            "dill==0.3.4",
        ],
        extras_require={"doc": ["nbsphinx", "recommonmark"]},
        python_requires=">=3.5",
        include_package_data=True,
        packages=find_packages(),
    )



# test the controlled unitary function


import torchquantum as tq
import torchquantum.functional as tqf

# import pdb
# pdb.set_trace()

flag = 8

if flag == 1:
    state = tq.QuantumState(n_wires=3)
    print(state)

    state.qubitunitaryfast(wires=0, params=[[0, 1], [1, 0]])
    state.qubitunitaryfast(wires=2, params=[[0, 1], [1, 0]])

    print(state)
    state.qubitunitaryfast(
        wires=[0, 2, 1],
        params=(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 1, 0],
            ]
        ),
    )

    print(state)

    state = tq.QuantumState(n_wires=3)
    print(state)

    gate1 = tq.QubitUnitaryFast(init_params=[[0, 1], [1, 0]], n_wires=1, wires=0)
    gate1(state)

    gate2 = tq.QubitUnitaryFast(init_params=[[0, 1], [1, 0]], n_wires=1, wires=2)

    gate2(state)
    print(state)

    gate3 = tq.QubitUnitaryFast(
        init_params=(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 1, 0],
            ]
        ),
        n_wires=3,
        wires=[0, 2, 1],
    )

    ## use controlled gate to implement a cnot

    gate3(state)
    print(state)

if flag == 2:
    state = tq.QuantumState(n_wires=2)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.PauliX(), c_wires=0, t_wires=1, trainable=False
    )

    gate0(state)
    print(state)

    gate(state)
    print(state)

if flag == 3:
    state = tq.QuantumState(n_wires=2)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate = tq.QubitUnitaryFast.fromControlledOperation(
        op=tq.RX(has_params=True, init_params=0.25),
        c_wires=0,
        t_wires=1,
        trainable=False,
    )

    gate0(state)
    print(state)

    gate(state)
    print(state)


if flag == 4:
    state = tq.QuantumState(n_wires=3)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate1 = tq.PauliX(n_wires=1, wires=1)
    gate = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.CNOT(), c_wires=0, t_wires=[1, 2], trainable=False
    )

    gate0(state)
    print(state)

    gate1(state)
    print(state)

    gate(state)
    print(state)


if flag == 5:
    state = tq.QuantumState(n_wires=5)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=1)
    gate1 = tq.PauliX(n_wires=1, wires=0)
    gate2 = tq.PauliX(n_wires=1, wires=4)
    gate = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.CNOT(), c_wires=1, t_wires=[[0, 2], [4, 3]], trainable=False
    )

    gate0(state)
    print(state)

    gate1(state)
    print(state)

    gate2(state)
    print(state)

    gate(state)
    print(state)


if flag == 6:
    state = tq.QuantumState(n_wires=5)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate1 = tq.PauliX(n_wires=1, wires=1)
    gate2 = tq.PauliX(n_wires=1, wires=2)
    gate3 = tq.PauliX(n_wires=1, wires=3)

    # gate2 = tq.PauliX(n_wires=1, wires=4)
    gate = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.Toffoli(), c_wires=[0, 1], t_wires=[2, 3, 4], trainable=False
    )

    gate0(state)
    print(state)

    # gate1(state)
    # print(state)

    gate2(state)
    print(state)

    gate3(state)
    print(state)

    gate(state)
    print(state)

if flag == 7:
    state = tq.QuantumState(n_wires=9)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate1 = tq.PauliX(n_wires=1, wires=1)
    gate2 = tq.PauliX(n_wires=1, wires=2)
    gate3 = tq.PauliX(n_wires=1, wires=3)
    gate4 = tq.PauliX(n_wires=1, wires=4)
    gate5 = tq.PauliX(n_wires=1, wires=6)
    gate6 = tq.PauliX(n_wires=1, wires=7)

    # gate2 = tq.PauliX(n_wires=1, wires=4)
    gate = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.Toffoli(),
        c_wires=[0, 1, 2],
        t_wires=[[3, 4, 5], [6, 7, 8]],
        trainable=False,
    )

    gate0(state)
    print(state)

    gate1(state)
    print(state)

    gate2(state)
    print(state)

    gate3(state)
    print(state)

    gate4(state)
    print(state)

    gate5(state)
    print(state)

    gate6(state)
    print(state)

    gate(state)
    print(state)


if flag == 8:
    state = tq.QuantumState(n_wires=3)
    print(state)
    gate0 = tq.PauliX(n_wires=1, wires=0)
    gate1 = tq.PauliX(n_wires=1, wires=1)
    gate_cx = tq.QubitUnitaryFast.from_controlled_operation(
        op=tq.PauliX(), c_wires=0, t_wires=1, trainable=False
    )

    gate_ccx = tq.QubitUnitaryFast.from_controlled_operation(
        op=gate_cx, c_wires=0, t_wires=[1, 2], trainable=False
    )

    gate0(state)
    print(state)

    gate1(state)
    print(state)

    gate_ccx(state)
    print(state)



import numpy as np
import torch
import torchquantum as tq

from torchpack.utils.logging import logger
from qiskit.providers.aer.noise import NoiseModel
from torchquantum.utils import get_provider


__all__ = [
    "NoiseModelTQ",
    "NoiseModelTQActivation",
    "NoiseModelTQPhase",
    "NoiseModelTQReadoutOnly",
    "NoiseModelTQActivationReadout",
    "NoiseModelTQPhaseReadout",
    "NoiseModelTQQErrorOnly",
]


def cos_adjust_noise(
    current_epoch,
    n_epochs,
    prob_schedule,
    prob_schedule_separator,
    orig_noise_total_prob,
):
    if prob_schedule is None:
        noise_total_prob = orig_noise_total_prob
    elif prob_schedule == "increase":
        # scale the cos
        if current_epoch <= prob_schedule_separator:
            noise_total_prob = orig_noise_total_prob * (
                -np.cos(current_epoch / prob_schedule_separator * np.pi) / 2 + 0.5
            )
        else:
            noise_total_prob = orig_noise_total_prob
    elif prob_schedule == "decrease":
        if current_epoch >= prob_schedule_separator:
            noise_total_prob = orig_noise_total_prob * (
                np.cos(
                    (current_epoch - prob_schedule_separator)
                    / (n_epochs - prob_schedule_separator)
                    * np.pi
                )
                / 2
                + 0.5
            )
        else:
            noise_total_prob = orig_noise_total_prob
    elif prob_schedule == "increase_decrease":
        # if current_epoch <= self.prob_schedule_separator:
        #     self.noise_total_prob = self.orig_noise_total_prob * \
        #         1 / (1 + np.exp(-(current_epoch - (
        #             self.prob_schedule_separator / 2)) / 10))
        # else:
        #     self.noise_total_prob = self.orig_noise_total_prob * \
        #         1 / (1 + np.exp((current_epoch - (
        #             self.n_epochs + self.prob_schedule_separator) / 2) /
        #                 10))
        if current_epoch <= prob_schedule_separator:
            noise_total_prob = orig_noise_total_prob * (
                -np.cos(current_epoch / prob_schedule_separator * np.pi) / 2 + 0.5
            )
        else:
            noise_total_prob = orig_noise_total_prob * (
                np.cos(
                    (current_epoch - prob_schedule_separator)
                    / (n_epochs - prob_schedule_separator)
                    * np.pi
                )
                / 2
                + 0.5
            )
    else:
        logger.warning(
            f"Not implemented schedule{prob_schedule}, " f"will not change prob!"
        )
        noise_total_prob = orig_noise_total_prob

    return noise_total_prob


def apply_readout_error_func(x, c2p_mapping, measure_info):
    # add readout error
    noise_free_0_probs = (x + 1) / 2
    noise_free_1_probs = 1 - (x + 1) / 2

    noisy_0_to_0_prob_all = []
    noisy_0_to_1_prob_all = []
    noisy_1_to_0_prob_all = []
    noisy_1_to_1_prob_all = []

    for k in range(x.shape[-1]):
        p_wire = [c2p_mapping[k]]
        noisy_0_to_0_prob_all.append(measure_info[tuple(p_wire)]["probabilities"][0][0])
        noisy_0_to_1_prob_all.append(measure_info[tuple(p_wire)]["probabilities"][0][1])
        noisy_1_to_0_prob_all.append(measure_info[tuple(p_wire)]["probabilities"][1][0])
        noisy_1_to_1_prob_all.append(measure_info[tuple(p_wire)]["probabilities"][1][1])

    noisy_0_to_0_prob_all = torch.tensor(noisy_0_to_0_prob_all, device=x.device)
    noisy_0_to_1_prob_all = torch.tensor(noisy_0_to_1_prob_all, device=x.device)
    noisy_1_to_0_prob_all = torch.tensor(noisy_1_to_0_prob_all, device=x.device)
    noisy_1_to_1_prob_all = torch.tensor(noisy_1_to_1_prob_all, device=x.device)

    noisy_measured_0 = (
        noise_free_0_probs * noisy_0_to_0_prob_all
        + noise_free_1_probs * noisy_1_to_0_prob_all
    )

    noisy_measured_1 = (
        noise_free_0_probs * noisy_0_to_1_prob_all
        + noise_free_1_probs * noisy_1_to_1_prob_all
    )
    noisy_expectation = noisy_measured_0 * 1 + noisy_measured_1 * (-1)

    return noisy_expectation


class NoiseModelTQ(object):
    """
    apply gate insertion and readout
    """

    def __init__(
        self,
        noise_model_name,
        n_epochs,
        noise_total_prob=None,
        ignored_ops=("id", "kraus", "reset"),
        prob_schedule=None,
        prob_schedule_separator=None,
        factor=None,
        add_thermal=True,
    ):
        self.noise_model_name = noise_model_name
        provider = get_provider(backend_name=noise_model_name)
        backend = provider.get_backend(noise_model_name)

        self.noise_model = NoiseModel.from_backend(
            backend, thermal_relaxation=add_thermal
        )
        self.noise_model_dict = self.noise_model.to_dict()
        self.is_add_noise = True
        self.v_c_reg_mapping = None
        self.p_c_reg_mapping = None
        self.p_v_reg_mapping = None
        self.orig_noise_total_prob = noise_total_prob
        self.noise_total_prob = noise_total_prob
        self.mode = "train"
        self.ignored_ops = ignored_ops

        self.parsed_dict = self.parse_noise_model_dict(self.noise_model_dict)
        self.parsed_dict = self.clean_parsed_noise_model_dict(
            self.parsed_dict, ignored_ops
        )
        self.n_epochs = n_epochs
        self.prob_schedule = prob_schedule
        self.prob_schedule_separator = prob_schedule_separator
        self.factor = factor

    def adjust_noise(self, current_epoch):
        self.noise_total_prob = cos_adjust_noise(
            current_epoch=current_epoch,
            n_epochs=self.n_epochs,
            prob_schedule=self.prob_schedule,
            prob_schedule_separator=self.prob_schedule_separator,
            orig_noise_total_prob=self.orig_noise_total_prob,
        )

    @staticmethod
    def clean_parsed_noise_model_dict(nm_dict, ignored_ops):
        # remove the ignored operation in the instructions and probs
        for operation, operation_info in nm_dict.items():
            for qubit, qubit_info in operation_info.items():
                inst_all = []
                prob_all = []
                if qubit_info["type"] == "qerror":
                    for inst, prob in zip(
                        qubit_info["instructions"], qubit_info["probabilities"]
                    ):
                        if any([inst_one["name"] in ignored_ops for inst_one in inst]):
                            continue
                        else:
                            inst_all.append(inst)
                            prob_all.append(prob)
                    nm_dict[operation][qubit]["instructions"] = inst_all
                    nm_dict[operation][qubit]["probabilities"] = prob_all
        return nm_dict

    @staticmethod
    def parse_noise_model_dict(nm_dict):
        # the qubits here are physical (p) qubits
        parsed = {}
        nm_list = nm_dict["errors"]

        for info in nm_list:
            val_dict = {
                "type": info["type"],
                "instructions": info.get("instructions", None),
                "probabilities": info["probabilities"],
            }

            if info["operations"][0] not in parsed.keys():
                parsed[info["operations"][0]] = {
                    tuple(info["gate_qubits"][0]): val_dict
                }
            elif (
                tuple(info["gate_qubits"][0])
                not in parsed[info["operations"][0]].keys()
            ):
                parsed[info["operations"][0]][tuple(info["gate_qubits"][0])] = val_dict
            else:
                raise ValueError

        return parsed

    def magnify_probs(self, probs):
        if self.factor is not None:
            factor = self.factor
        else:
            if self.noise_total_prob is None:
                factor = 1
            else:
                factor = self.noise_total_prob / sum(probs)
        probs = [prob * factor for prob in probs]

        return probs

    def sample_noise_op(self, op_in):
        if not (self.mode == "train" and self.is_add_noise):
            return []

        op_name = op_in.name.lower()
        if op_name == "paulix":
            op_name = "x"
        elif op_name == "cnot":
            op_name = "cx"
        elif op_name in ["sx", "id"]:
            pass
        elif op_name == "rz":
            # no noise
            return []
        else:
            logger.warning(f"No noise model for {op_name} operation!")

        wires = op_in.wires

        p_wires = [self.p_v_reg_mapping["v2p"][wire] for wire in wires]

        if tuple(p_wires) in self.parsed_dict[op_name].keys():
            inst_prob = self.parsed_dict[op_name][tuple(p_wires)]
        else:
            # not in the real coupling map, so only give a dummy one
            if len(p_wires) == 1:
                inst_prob = self.parsed_dict[op_name][(0,)]
            elif len(p_wires) == 2:
                inst_prob = self.parsed_dict[op_name][(0, 1)]

        inst = inst_prob["instructions"]
        if len(inst) == 0:
            return []

        probs = inst_prob["probabilities"]

        magnified_probs = self.magnify_probs(probs)

        idx = np.random.choice(
            list(range(len(inst) + 1)), p=magnified_probs + [1 - sum(magnified_probs)]
        )
        if idx == len(inst):
            return []

        instructions = inst[idx]

        ops = []
        for instruction in instructions:
            v_wires = [
                self.p_v_reg_mapping["p2v"][qubit] for qubit in instruction["qubits"]
            ]
            if instruction["name"] == "x":
                op = tq.PauliX(wires=v_wires)
            elif instruction["name"] == "y":
                op = tq.PauliY(wires=v_wires)
            elif instruction["name"] == "z":
                op = tq.PauliZ(wires=v_wires)
            elif instruction["name"] == "reset":
                op = tq.Reset(wires=v_wires)
            else:
                # ignore operations specified by self.ignored_ops
                # logger.warning(f"skip noise operation {instruction['name']}")
                continue
            ops.append(op)

        return ops

    def apply_readout_error(self, x):
        c2p_mapping = self.p_c_reg_mapping["c2p"]
        measure_info = self.parsed_dict["measure"]

        return apply_readout_error_func(x, c2p_mapping, measure_info)


class NoiseModelTQActivation(object):
    """
    add noise to the activations
    """

    def __init__(
        self,
        mean=(0.0,),
        std=(1.0,),
        n_epochs=200,
        prob_schedule=None,
        prob_schedule_separator=None,
        after_norm=False,
        factor=None,
    ):
        self.mean = mean
        self.std = std
        self.is_add_noise = True
        self.mode = "train"
        self.after_norm = after_norm

        self.orig_std = std
        self.n_epochs = n_epochs
        self.prob_schedule = prob_schedule
        self.prob_schedule_separator = prob_schedule_separator
        self.factor = factor

    @property
    def noise_total_prob(self):
        return self.std

    @noise_total_prob.setter
    def noise_total_prob(self, value):
        self.std = value

    def adjust_noise(self, current_epoch):
        self.std = cos_adjust_noise(
            current_epoch=current_epoch,
            n_epochs=self.n_epochs,
            prob_schedule=self.prob_schedule,
            prob_schedule_separator=self.prob_schedule_separator,
            orig_noise_total_prob=self.orig_std,
        )

    def sample_noise_op(self, op_in):
        return []

    def apply_readout_error(self, x):
        return x

    def add_noise(self, x, node_id, is_after_norm=False):
        if (self.after_norm and is_after_norm) or (
            not self.after_norm and not is_after_norm
        ):
            if self.mode == "train" and self.is_add_noise:
                if self.factor is None:
                    factor = 1
                else:
                    factor = self.factor

                x = (
                    x
                    + torch.randn(x.shape, device=x.device) * self.std[node_id] * factor
                    + self.mean[node_id]
                )

        return x


class NoiseModelTQPhase(object):
    """
    add noise to rotation parameters
    """

    def __init__(
        self,
        mean=0.0,
        std=1.0,
        n_epochs=200,
        prob_schedule=None,
        prob_schedule_separator=None,
        factor=None,
    ):
        self.mean = mean
        self.std = std
        self.is_add_noise = True
        self.mode = "train"

        self.orig_std = std
        self.n_epochs = n_epochs
        self.prob_schedule = prob_schedule
        self.prob_schedule_separator = prob_schedule_separator
        self.factor = factor

    @property
    def noise_total_prob(self):
        return self.std

    @noise_total_prob.setter
    def noise_total_prob(self, value):
        self.std = value

    def adjust_noise(self, current_epoch):
        self.std = cos_adjust_noise(
            current_epoch=current_epoch,
            n_epochs=self.n_epochs,
            prob_schedule=self.prob_schedule,
            prob_schedule_separator=self.prob_schedule_separator,
            orig_noise_total_prob=self.orig_std,
        )

    def sample_noise_op(self, op_in):
        return []

    def apply_readout_error(self, x):
        return x

    def add_noise(self, phase):
        if self.mode == "train" and self.is_add_noise:
            if self.factor is None:
                factor = 1
            else:
                factor = self.factor
            phase = (
                phase
                + torch.randn(phase.shape, device=phase.device) * self.std * factor
                + self.mean
            )

        return phase


class NoiseModelTQReadoutOnly(NoiseModelTQ):
    def sample_noise_op(self, op_in):
        return []


class NoiseModelTQQErrorOnly(NoiseModelTQ):
    def apply_readout_error(self, x):
        return x


class NoiseModelTQActivationReadout(NoiseModelTQActivation):
    def __init__(
        self,
        noise_model_name,
        mean=0.0,
        std=1.0,
        n_epochs=200,
        prob_schedule=None,
        prob_schedule_separator=None,
        after_norm=False,
        factor=None,
    ):
        super().__init__(
            mean=mean,
            std=std,
            n_epochs=n_epochs,
            prob_schedule=prob_schedule,
            prob_schedule_separator=prob_schedule_separator,
            after_norm=after_norm,
            factor=factor,
        )
        provider = get_provider(backend_name=noise_model_name)
        backend = provider.get_backend(noise_model_name)

        self.noise_model = NoiseModel.from_backend(backend)
        self.noise_model_dict = self.noise_model.to_dict()
        self.is_add_noise = True
        self.v_c_reg_mapping = None
        self.p_c_reg_mapping = None
        self.p_v_reg_mapping = None

        self.parsed_dict = NoiseModelTQ.parse_noise_model_dict(self.noise_model_dict)

    def apply_readout_error(self, x):
        c2p_mapping = self.p_c_reg_mapping["c2p"]
        measure_info = self.parsed_dict["measure"]

        return apply_readout_error_func(x, c2p_mapping, measure_info)


class NoiseModelTQPhaseReadout(NoiseModelTQPhase):
    def __init__(
        self,
        noise_model_name,
        mean=0.0,
        std=1.0,
        n_epochs=200,
        prob_schedule=None,
        prob_schedule_separator=None,
        factor=None,
    ):
        super().__init__(
            mean=mean,
            std=std,
            n_epochs=n_epochs,
            prob_schedule=prob_schedule,
            prob_schedule_separator=prob_schedule_separator,
            factor=factor,
        )
        provider = get_provider(backend_name=noise_model_name)
        backend = provider.get_backend(noise_model_name)

        self.noise_model = NoiseModel.from_backend(backend)
        self.noise_model_dict = self.noise_model.to_dict()
        self.is_add_noise = True
        self.v_c_reg_mapping = None
        self.p_c_reg_mapping = None
        self.p_v_reg_mapping = None

        self.parsed_dict = NoiseModelTQ.parse_noise_model_dict(self.noise_model_dict)

    def apply_readout_error(self, x):
        c2p_mapping = self.p_c_reg_mapping["c2p"]
        measure_info = self.parsed_dict["measure"]

        return apply_readout_error_func(x, c2p_mapping, measure_info)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Arbitrary unitary circuit instruction.
"""

from collections import OrderedDict
import numpy

from qiskit.circuit import Gate, ControlledGate
from qiskit.circuit import QuantumCircuit
from qiskit.circuit import QuantumRegister, Qubit
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit._utils import _compute_control_matrix
from qiskit.circuit.library.standard_gates import U3Gate
from qiskit.extensions.quantum_initializer import isometry
from qiskit.quantum_info.operators.predicates import matrix_equal
from qiskit.quantum_info.operators.predicates import is_unitary_matrix
from qiskit.quantum_info.synthesis.one_qubit_decompose import OneQubitEulerDecomposer
from qiskit.quantum_info.synthesis.two_qubit_decompose import two_qubit_cnot_decompose
from qiskit.extensions.exceptions import ExtensionError

_DECOMPOSER1Q = OneQubitEulerDecomposer("U3")


class UnitaryGate(Gate):
    """Class for representing unitary gates"""

    def __init__(self, data, label=None):
        """Create a gate from a numeric unitary matrix.

        Args:
            data (matrix or Operator): unitary operator.
            label (str): unitary name for backend [Default: None].

        Raises:
            ExtensionError: if input data is not an N-qubit unitary operator.
        """
        if hasattr(data, "to_matrix"):
            # If input is Gate subclass or some other class object that has
            # a to_matrix method this will call that method.
            data = data.to_matrix()
        elif hasattr(data, "to_operator"):
            # If input is a BaseOperator subclass this attempts to convert
            # the object to an Operator so that we can extract the underlying
            # numpy matrix from `Operator.data`.
            data = data.to_operator().data
        # Convert to numpy array in case not already an array
        data = numpy.array(data, dtype=complex)
        # Check input is unitary
        if not is_unitary_matrix(data, atol=1e-5):
            raise ExtensionError("Input matrix is not unitary.")
        # Check input is N-qubit matrix
        input_dim, output_dim = data.shape
        num_qubits = int(numpy.log2(input_dim))
        if input_dim != output_dim or 2**num_qubits != input_dim:
            raise ExtensionError("Input matrix is not an N-qubit operator.")

        self._qasm_name = None
        self._qasm_definition = None
        self._qasm_def_written = False
        # Store instruction params
        super().__init__("unitary", num_qubits, [data], label=label)

    def __eq__(self, other):
        if not isinstance(other, UnitaryGate):
            return False
        if self.label != other.label:
            return False
        # Should we match unitaries as equal if they are equal
        # up to global phase?
        return matrix_equal(self.params[0], other.params[0], ignore_phase=True)

    def to_matrix(self):
        """Return matrix for the unitary."""
        return self.params[0]

    def inverse(self):
        """Return the adjoint of the unitary."""
        return self.adjoint()

    def conjugate(self):
        """Return the conjugate of the unitary."""
        return UnitaryGate(numpy.conj(self.to_matrix()))

    def adjoint(self):
        """Return the adjoint of the unitary."""
        return self.transpose().conjugate()

    def transpose(self):
        """Return the transpose of the unitary."""
        return UnitaryGate(numpy.transpose(self.to_matrix()))

    def _define(self):
        """Calculate a subcircuit that implements this unitary."""
        if self.num_qubits == 1:
            q = QuantumRegister(1, "q")
            qc = QuantumCircuit(q, name=self.name)
            theta, phi, lam, global_phase = _DECOMPOSER1Q.angles_and_phase(
                self.to_matrix()
            )
            qc._append(U3Gate(theta, phi, lam), [q[0]], [])
            qc.global_phase = global_phase
            self.definition = qc
        elif self.num_qubits == 2:
            self.definition = two_qubit_cnot_decompose(self.to_matrix())
        else:
            q = QuantumRegister(self.num_qubits, "q")
            qc = QuantumCircuit(q, name=self.name)
            qc.append(isometry.Isometry(self.to_matrix(), 0, 0), qargs=q[:])
            self.definition = qc

    def control(self, num_ctrl_qubits=1, label=None, ctrl_state=None):
        r"""Return controlled version of gate

        Args:
            num_ctrl_qubits (int): number of controls to add to gate (default=1)
            label (str): optional gate label
            ctrl_state (int or str or None): The control state in decimal or as a
                bit string (e.g. '1011'). If None, use 2**num_ctrl_qubits-1.

        Returns:
            UnitaryGate: controlled version of gate.

        Raises:
            QiskitError: Invalid ctrl_state.
            ExtensionError: Non-unitary controlled unitary.
        """
        cmat = _compute_control_matrix(
            self.to_matrix(), num_ctrl_qubits, ctrl_state=None
        )
        iso = isometry.Isometry(cmat, 0, 0)
        cunitary = ControlledGate(
            "c-unitary",
            num_qubits=self.num_qubits + num_ctrl_qubits,
            params=[cmat],
            label=label,
            num_ctrl_qubits=num_ctrl_qubits,
            definition=iso.definition,
            ctrl_state=ctrl_state,
            base_gate=self.copy(),
        )
        from qiskit.quantum_info import Operator

        # hack to correct global phase; should fix to prevent need for correction here
        pmat = Operator(iso.inverse()).data @ cmat
        diag = numpy.diag(pmat)
        if not numpy.allclose(diag, diag[0]):
            raise ExtensionError("controlled unitary generation failed")
        phase = numpy.angle(diag[0])
        if phase:
            # need to apply to _definition since open controls creates temporary definition
            cunitary._definition.global_phase = phase
        return cunitary

    def qasm(self):
        """The qasm for a custom unitary gate
        This is achieved by adding a custom gate that corresponds to the definition
        of this gate. It gives the gate a random name if one hasn't been given to it.
        """
        # if this is true then we have written the gate definition already
        # so we only need to write the name
        if self._qasm_def_written:
            return self._qasmif(self._qasm_name)

        # we have worked out the definition before, but haven't written it yet
        # so we need to write definition + name
        if self._qasm_definition:
            self._qasm_def_written = True
            return self._qasm_definition + self._qasmif(self._qasm_name)

        # need to work out the definition and then write it

        # give this unitary a name
        self._qasm_name = self.label if self.label else "unitary" + str(id(self))

        # map from gates in the definition to params in the method
        reg_to_qasm = OrderedDict()
        current_reg = 0

        gates_def = ""
        for gate in self.definition.data:

            # add regs from this gate to the overall set of params
            for reg in gate[1] + gate[2]:
                if reg not in reg_to_qasm:
                    reg_to_qasm[reg] = "p" + str(current_reg)
                    current_reg += 1

            curr_gate = "\t%s %s;\n" % (
                gate[0].qasm(),
                ",".join([reg_to_qasm[j] for j in gate[1] + gate[2]]),
            )
            gates_def += curr_gate

        # name of gate + params + {definition}
        overall = (
            "gate "
            + self._qasm_name
            + " "
            + ",".join(reg_to_qasm.values())
            + " {\n"
            + gates_def
            + "}\n"
        )

        self._qasm_def_written = True
        self._qasm_definition = overall

        return self._qasm_definition + self._qasmif(self._qasm_name)

    def validate_parameter(self, parameter):
        """Unitary gate parameter has to be an ndarray."""
        if isinstance(parameter, numpy.ndarray):
            return parameter
        else:
            raise CircuitError(
                "invalid param type {0} in gate "
                "{1}".format(type(parameter), self.name)
            )


def unitary(self, obj, qubits, label=None):
    """Apply unitary gate to q."""
    gate = UnitaryGate(obj, label=label)
    if isinstance(qubits, QuantumRegister):
        qubits = qubits[:]
    # for single qubit unitary gate, allow an 'int' or a 'list of ints' as qubits.
    if gate.num_qubits == 1:
        if isinstance(qubits, (int, Qubit)) or len(qubits) > 1:
            qubits = [qubits]
    return self.append(gate, qubits, [])


QuantumCircuit.unitary = unitary



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""A module for monitoring various qiskit functionality"""

import sys
import time
from datetime import datetime, timezone


def _text_checker(
    job,
    interval,
    _interval_set=False,
    quiet=False,
    output=sys.stdout,
    line_discipline="\r",
):
    """A text-based job status checker
    Args:
        job (BaseJob): The job to check.
        interval (int): The interval at which to check.
        _interval_set (bool): Was interval time set by user?
        quiet (bool): If True, do not print status messages.
        output (file): The file like object to write status messages to.
        By default this is sys.stdout.
        line_discipline (string): character emitted at start of a line of job monitor output,
        This defaults to \\r.
    """
    status = job.status()
    msg = status.value
    prev_msg = msg
    msg_len = len(msg)

    if not quiet:
        print(
            "{}{}: {}".format(line_discipline, "Job Status", msg), end="", file=output
        )
    while status.name not in ["DONE", "CANCELLED", "ERROR"]:
        time.sleep(interval)
        status = job.status()
        msg = status.value

        if status.name == "QUEUED":
            msg += " (%s)" % job.queue_position()
            if job.queue_position() is None:
                interval = 2
            elif not _interval_set:
                interval = max(job.queue_position(), 2)
        else:
            if not _interval_set:
                interval = 2

        # Adjust length of message so there are no artifacts
        if len(msg) < msg_len:
            msg += " " * (msg_len - len(msg))
        elif len(msg) > msg_len:
            msg_len = len(msg)

        queue_info = job.queue_info()
        if queue_info is not None:
            now = datetime.now(timezone.utc)
            job_id = queue_info.job_id
            wait_time = (queue_info.estimated_start_time - now).seconds / 60
            msg += f"\t est. wait {wait_time:.3f} min"
            msg += f"\t priority: {queue_info.hub_priority}(hub), {queue_info.group_priority}(group), {queue_info.project_priority}(proj)"
            msg += f"\t job id: {job_id}"

        if msg != prev_msg and not quiet:
            print(
                "{}{}: {}".format(line_discipline, "Job Status", msg),
                end="",
                file=output,
            )
            prev_msg = msg

    if not quiet:
        print("", file=output)


def my_job_monitor(
    job, interval=None, quiet=False, output=sys.stdout, line_discipline="\r"
):
    """Monitor the status of a IBMQJob instance.
    Args:
        job (BaseJob): Job to monitor.
        interval (int): Time interval between status queries.
        quiet (bool): If True, do not print status messages.
        output (file): The file like object to write status messages to.
        By default this is sys.stdout.
        line_discipline (string): character emitted at start of a line of job monitor output,
        This defaults to \\r.
    """
    if interval is None:
        _interval_set = False
        interval = 5
    else:
        _interval_set = True

    _text_checker(
        job,
        interval,
        _interval_set,
        quiet=quiet,
        output=output,
        line_discipline=line_discipline,
    )



from torchpack.datasets.dataset import Dataset


__all__ = ["VQE"]


class VQEDataset:
    def __init__(self, split, steps_per_epoch):
        self.split = split
        self.steps_per_epoch = steps_per_epoch

    def __getitem__(self, index: int):
        instance = {"input": -1, "target": -1}

        return instance

    def __len__(self) -> int:
        if self.split == "train":
            return self.steps_per_epoch
        else:
            return 1


class VQE(Dataset):
    def __init__(self, steps_per_epoch):
        super().__init__(
            {
                split: VQEDataset(split=split, steps_per_epoch=steps_per_epoch)
                for split in ["train", "valid", "test"]
            }
        )



import torch
import torchquantum as tq

from torchquantum.encoding import encoder_op_list_name_dict
from torchquantum.layers import layer_name_dict
import numpy as np
import datetime


__all__ = [
    "QuantumNode",
    "build_nodes",
]


class QuantumNode(tq.QuantumModule):
    """
    a quantum node contains a q device, encoder, q layer and measure
    """

    def __init__(self, arch, act_norm, node_id):
        super().__init__()
        self.arch = arch
        self.q_device = tq.QuantumDevice(n_wires=arch["n_wires"])
        self.encoder = tq.GeneralEncoder(
            encoder_op_list_name_dict[arch["encoder_op_list_name"]]
        )
        self.q_layer = layer_name_dict[arch["q_layer_name"]](arch)
        self.measure = tq.MeasureAll(tq.PauliZ)
        self.act_norm = act_norm
        self.x_before_add_noise = None
        self.x_before_add_noise_second = None
        self.x_before_act_quant = None
        self.x_before_norm = None
        self.circuit_in = None
        self.circuit_out = None
        self.shift_this_step = np.array([True] * len(list(self.q_layer.parameters())))
        self.cool_down = [0] * len(list(self.q_layer.parameters()))
        self.triger_cd = [0] * len(list(self.q_layer.parameters()))
        if self.act_norm == "batch_norm" or self.act_norm == "batch_norm_no_last":
            self.bn = torch.nn.BatchNorm1d(
                num_features=arch["n_wires"],
                momentum=None,
                affine=False,
                track_running_stats=False,
            )
        self.node_id = node_id
        self.pre_specified_mean_std = None
        self.grad_qlayer = None
        self.grad_encoder = None

    def forward(self, x, use_qiskit=False, is_last_node=False, parallel=True):
        if use_qiskit:
            x = self.qiskit_processor.process_parameterized(
                self.q_device,
                self.encoder,
                self.q_layer,
                self.measure,
                x,
                parallel=parallel,
            )
        else:
            self.encoder(self.q_device, x)
            self.q_layer(self.q_device)
            x = self.measure(self.q_device)

        self.x_before_add_noise = x.clone()

        if isinstance(self.noise_model_tq, tq.NoiseModelTQActivation):
            x = self.noise_model_tq.add_noise(x, self.node_id, is_after_norm=False)

        self.x_before_norm = x.clone()

        if self.act_norm == "layer_norm":
            x = (x - x.mean(-1).unsqueeze(-1)) / x.std(-1).unsqueeze(-1)
        elif self.act_norm == "batch_norm":
            if self.pre_specified_mean_std is None:
                x = self.bn(x)
            else:
                x = (
                    x
                    - torch.tensor(
                        self.pre_specified_mean_std["mean"], device=x.device
                    ).unsqueeze(0)
                ) / torch.tensor(
                    self.pre_specified_mean_std["std"], device=x.device
                ).unsqueeze(
                    0
                )

            # x = (x - x.mean(0).unsqueeze(0)) / x.std(0).unsqueeze(0)
        elif self.act_norm == "all_norm":
            x = (x - x.mean()) / x.std()
        elif self.act_norm == "layer_norm_no_last":
            if not is_last_node:
                x = (x - x.mean(-1).unsqueeze(-1)) / x.std(-1).unsqueeze(-1)
        elif self.act_norm == "batch_norm_no_last":

            if not is_last_node:
                if self.pre_specified_mean_std is None:
                    x = self.bn(x)
                else:
                    x = (
                        x
                        - torch.tensor(
                            self.pre_specified_mean_std["mean"], device=x.device
                        ).unsqueeze(0)
                    ) / torch.tensor(
                        self.pre_specified_mean_std["std"], device=x.device
                    ).unsqueeze(
                        0
                    )

        self.x_before_add_noise_second = x.clone()

        if isinstance(self.noise_model_tq, tq.NoiseModelTQActivation):
            x = self.noise_model_tq.add_noise(x, self.node_id, is_after_norm=True)

        self.x_before_act_quant = x.clone()

        return x

    def run_circuit(self, inputs):
        self.encoder(self.q_device, inputs)
        self.q_layer(self.q_device)
        x = self.measure(self.q_device)
        return x

    def shift_and_run(
        self,
        x,
        use_qiskit=False,
        is_last_node=False,
        is_first_node=False,
        parallel=True,
    ):
        import numpy as np

        self.circuit_in = x
        self.circuit_out = None
        time_spent = datetime.timedelta()
        if use_qiskit:
            with torch.no_grad():
                bsz = x.shape[0]
                inputs = x
                (
                    x,
                    time_spent_list,
                ) = self.qiskit_processor.process_parameterized_and_shift(
                    self.q_device,
                    self.encoder,
                    self.q_layer,
                    self.measure,
                    inputs,
                    shift_encoder=False,
                    parallel=parallel,
                    shift_this_step=self.shift_this_step,
                )
                for ts in time_spent_list:
                    time_spent = time_spent + ts
                results = x.reshape(
                    1 + 2 * np.sum(self.shift_this_step), bsz, self.arch["n_wires"]
                )
                self.circuit_out = results[0, :, :].clone()

                cnt = 0
                self.grad_qlayer = []
                for i, named_param in enumerate(self.q_layer.named_parameters()):
                    if self.shift_this_step[i]:
                        cnt = cnt + 1
                        out1 = results[cnt, :, :]
                        cnt = cnt + 1
                        out2 = results[cnt, :, :]
                        self.grad_qlayer.append(0.5 * (out1 - out2))
                    else:
                        self.grad_qlayer.append(None)

                self.grad_encoder = []
                if not is_first_node:
                    (
                        x,
                        time_spent_list,
                    ) = self.qiskit_processor.process_parameterized_and_shift(
                        self.q_device,
                        self.encoder,
                        self.q_layer,
                        self.measure,
                        inputs,
                        shift_encoder=True,
                        parallel=parallel,
                    )
                    for ts in time_spent_list:
                        time_spent = time_spent + ts
                    results = x.reshape(2 * inputs.shape[1], bsz, self.arch["n_wires"])
                    cnt = 0
                    while cnt < 2 * inputs.shape[1]:
                        out1 = results[cnt, :, :]
                        cnt = cnt + 1
                        out2 = results[cnt, :, :]
                        cnt = cnt + 1
                        self.grad_encoder.append(0.5 * (out1 - out2))
        else:
            with torch.no_grad():
                # time1 = datetime.datetime.now()
                inputs = x
                x = self.run_circuit(inputs)
                self.circuit_out = x

                self.grad_qlayer = []
                for i, param in enumerate(self.q_layer.parameters()):
                    if self.shift_this_step[i]:
                        param.copy_(param + np.pi * 0.5)
                        out1 = self.run_circuit(inputs)
                        param.copy_(param - np.pi)
                        out2 = self.run_circuit(inputs)
                        param.copy_(param + np.pi * 0.5)
                        self.grad_qlayer.append(0.5 * (out1 - out2))
                    else:
                        self.grad_qlayer.append(None)

                self.grad_encoder = []
                if not is_first_node:
                    for input_id in range(inputs.size()[1]):
                        inputs[:, input_id] += np.pi * 0.5
                        out1 = self.run_circuit(inputs)
                        inputs[:, input_id] -= np.pi
                        out2 = self.run_circuit(inputs)
                        inputs[:, input_id] += np.pi * 0.5
                        self.grad_encoder.append(0.5 * (out1 - out2))

                # time2 = datetime.datetime.now()
                # print('one step:')
                # print(time2 -time1)
                # print('run one circuit:')
                # print((time2 -time1) / (1 + 2 * np.sum(self.shift_this_step)))

        x = self.circuit_out
        self.x_before_add_noise = x.clone()
        self.circuit_out.requires_grad = True

        if isinstance(self.noise_model_tq, tq.NoiseModelTQActivation):
            x = self.noise_model_tq.add_noise(x, self.node_id, is_after_norm=False)

        self.x_before_norm = x.clone()

        if self.act_norm == "layer_norm":
            x = (x - x.mean(-1).unsqueeze(-1)) / x.std(-1).unsqueeze(-1)
        elif self.act_norm == "batch_norm":
            if self.pre_specified_mean_std is None:
                x = self.bn(x)
            else:
                x = (
                    x
                    - torch.tensor(
                        self.pre_specified_mean_std["mean"], device=x.device
                    ).unsqueeze(0)
                ) / torch.tensor(
                    self.pre_specified_mean_std["std"], device=x.device
                ).unsqueeze(
                    0
                )

            # x = (x - x.mean(0).unsqueeze(0)) / x.std(0).unsqueeze(0)
        elif self.act_norm == "all_norm":
            x = (x - x.mean()) / x.std()
        elif self.act_norm == "layer_norm_no_last":
            if not is_last_node:
                x = (x - x.mean(-1).unsqueeze(-1)) / x.std(-1).unsqueeze(-1)
        elif self.act_norm == "batch_norm_no_last":

            if not is_last_node:
                if self.pre_specified_mean_std is None:
                    x = self.bn(x)
                else:
                    x = (
                        x
                        - torch.tensor(
                            self.pre_specified_mean_std["mean"], device=x.device
                        ).unsqueeze(0)
                    ) / torch.tensor(
                        self.pre_specified_mean_std["std"], device=x.device
                    ).unsqueeze(
                        0
                    )

        self.x_before_add_noise_second = x.clone()

        if isinstance(self.noise_model_tq, tq.NoiseModelTQActivation):
            x = self.noise_model_tq.add_noise(x, self.node_id, is_after_norm=True)

        self.x_before_act_quant = x.clone()

        return x, time_spent


def build_nodes(node_archs, act_norm=None):
    nodes = tq.QuantumModuleList()
    for k, node_arch in enumerate(node_archs):
        nodes.append(QuantumNode(node_arch, act_norm=act_norm, node_id=k))

    return nodes



import torchquantum as tq
import torch
import torch.nn.functional as F
import numpy as np
import os

from torchpack.utils.logging import logger
from torchpack.utils.config import configs


class QMultiFCModel0(tq.QuantumModule):
    # multiple nodes, one node contains encoder, q_layer, and measure
    def __init__(self, arch):
        super().__init__()
        self.arch = arch
        self.n_nodes = arch["n_nodes"]
        self.nodes = tq.build_nodes(arch["node_archs"], act_norm=arch["act_norm"])
        assert arch["n_nodes"] == len(arch["node_archs"])
        self.mse_all = []
        self.residual = getattr(arch, "residual", False)
        self.activations = []
        self.work_from_step = 0
        self.grad_dict = None
        self.count1 = 0
        self.count2 = 0
        self.num_forwards = 0
        self.n_params = len(list(self.nodes[0].parameters()))
        self.last_abs_grad = torch.zeros(self.n_params)
        self.pruning_method = arch["pruning_method"]
        if self.pruning_method == "random_pruning":
            self.sampling_ratio = 1 - arch["pruning_ratio"]
        elif self.pruning_method == "perlayer_pruning":
            self.n_qubits = arch["node_archs"][0]["n_wires"]
            self.n_layers = (
                arch["node_archs"][0]["n_layers_per_block"]
                * arch["node_archs"][0]["n_blocks"]
            )
            self.n_sampling_layers = self.n_layers - arch["n_pruning_layers"]
            self.colums = np.arange(self.n_sampling_layers)
        elif self.pruning_method == "perqubit_pruning":
            self.n_qubits = arch["node_archs"][0]["n_wires"]
            self.n_layers = (
                arch["node_archs"][0]["n_layers_per_block"]
                * arch["node_archs"][0]["n_blocks"]
            )
            self.n_sampling_qubits = self.n_qubits - arch["n_pruning_qubits"]
            self.rows = np.arange(self.n_sampling_qubits)
        elif self.pruning_method == "gradient_based_pruning":
            self.accumulation_window_size = arch["accumulation_window_size"]
            self.pruning_window_size = arch["pruning_window_size"]
            self.sampling_ratio = 1 - arch["pruning_ratio"]
            self.sum_abs_grad = torch.tensor([0.01] * self.n_params)
            self.is_accumulation = True
            self.accumulation_steps = 0
            self.pruning_steps = 0
        elif self.pruning_method == "gradient_based_deterministic":
            self.accumulation_window_size = arch["accumulation_window_size"]
            self.pruning_window_size = arch["pruning_window_size"]
            self.sampling_ratio = 1 - arch["pruning_ratio"]
            self.sum_abs_grad = torch.tensor([0.01] * self.n_params)
            self.is_accumulation = True
            self.accumulation_steps = 0
            self.pruning_steps = 0
        elif self.pruning_method == "phase_based_pruning":
            self.accumulation_window_size = arch["accumulation_window_size"]
            self.pruning_window_size = arch["pruning_window_size"]
            self.sampling_ratio = 1 - arch["pruning_ratio"]
            self.last_abs_param = torch.zeros(self.n_params)
            self.is_accumulation = True
            self.accumulation_steps = 0
            self.pruning_steps = 0
        else:
            logger.info("Not use any pruning")

    def forward(self, x, verbose=False, use_qiskit=False):
        bsz = x.shape[0]

        if getattr(self.arch, "down_sample_kernel_size", None) is not None:
            x = F.avg_pool2d(x, self.arch["down_sample_kernel_size"])

        if getattr(self.arch, "fft_remain_size", None) is not None:
            x = torch.fft.fft2(x, norm="ortho").abs()[
                :, :, : self.arch["fft_remain_size"], : self.arch["fft_remain_size"]
            ]
            x = x.contiguous()

        x = x.view(bsz, -1)
        mse_all = []

        for k, node in enumerate(self.nodes):
            node_out = node(
                x, use_qiskit=use_qiskit, is_last_node=(k == self.n_nodes - 1)
            )
            x = node_out
        self.mse_all = mse_all

        if getattr(self.arch, "output_len", None) is not None:
            x = x.reshape(bsz, -1, self.arch.output_len).sum(-1)

        if x.dim() > 2:
            x = x.squeeze()

        x = F.log_softmax(x, dim=1)
        return x

    def shift_and_run(
        self, x, global_step, total_step, verbose=False, use_qiskit=False
    ):
        bsz = x.shape[0]

        if getattr(self.arch, "down_sample_kernel_size", None) is not None:
            x = F.avg_pool2d(x, self.arch["down_sample_kernel_size"])

        if getattr(self.arch, "fft_remain_size", None) is not None:
            x = torch.fft.fft2(x, norm="ortho").abs()[
                :, :, : self.arch["fft_remain_size"], : self.arch["fft_remain_size"]
            ]
            x = x.contiguous()

        x = x.view(bsz, -1)
        mse_all = []

        for k, node in enumerate(self.nodes):
            node.shift_this_step[:] = True
            if self.pruning_method == "random_pruning":
                node.shift_this_step[:] = False
                idx = torch.randperm(self.n_params)[
                    : int(self.sampling_ratio * self.n_params)
                ]
                node.shift_this_step[idx] = True
            elif self.pruning_method == "perlayer_pruning":
                node.shift_this_step[:] = False
                idxs = torch.range(0, self.n_params - 1, dtype=int).view(
                    self.n_qubits, self.n_layers
                )
                sampled_colums = self.colums
                for colum in sampled_colums:
                    node.shift_this_step[idxs[:, colum]] = True
                self.colums += self.n_sampling_layers
                self.colums %= self.n_layers
            elif self.pruning_method == "perqubit_pruning":
                node.shift_this_step[:] = False
                idxs = torch.range(0, self.n_params - 1, dtype=int).view(
                    self.n_qubits, self.n_layers
                )
                sampled_rows = self.rows
                for row in sampled_rows:
                    node.shift_this_step[idxs[row]] = True
                self.rows += self.n_sampling_qubits
                self.rows %= self.n_qubits
            elif self.pruning_method == "gradient_based_pruning":
                if self.is_accumulation:
                    self.accumulation_steps += 1
                    self.sum_abs_grad = self.sum_abs_grad + self.last_abs_grad
                    node.shift_this_step[:] = True
                    if self.accumulation_steps == self.accumulation_window_size:
                        self.is_accumulation = False
                        self.accumulation_steps = 0
                        self.sum_abs_grad = torch.tensor([0.01] * self.n_params)
                else:
                    self.pruning_steps += 1
                    node.shift_this_step[:] = False
                    idx = torch.multinomial(
                        self.sum_abs_grad, int(self.sampling_ratio * self.n_params)
                    )
                    node.shift_this_step[idx] = True
                    if self.pruning_steps == self.pruning_window_size:
                        self.is_accumulation = True
                        self.pruning_steps = 0
            elif self.pruning_method == "gradient_based_deterministic":
                if self.is_accumulation:
                    self.accumulation_steps += 1
                    self.sum_abs_grad = self.sum_abs_grad + self.last_abs_grad
                    node.shift_this_step[:] = True
                    if self.accumulation_steps == self.accumulation_window_size:
                        self.is_accumulation = False
                        self.accumulation_steps = 0
                else:
                    self.pruning_steps += 1
                    node.shift_this_step[:] = False
                    idx = torch.argsort(self.sum_abs_grad, descending=True)[
                        : int(self.sampling_ratio * self.n_params)
                    ]
                    # idx = torch.multinomial(self.sum_abs_grad, int(self.sampling_ratio * self.n_params))
                    node.shift_this_step[idx] = True
                    if self.pruning_steps == self.pruning_window_size:
                        self.is_accumulation = True
                        self.pruning_steps = 0
                        self.sum_abs_grad = torch.tensor([0.01] * self.n_params)
            elif self.pruning_method == "phase_based_pruning":
                if self.is_accumulation:
                    self.accumulation_steps += 1
                    node.shift_this_step[:] = True
                    if self.accumulation_steps == self.accumulation_window_size:
                        self.is_accumulation = False
                        self.accumulation_steps = 0
                else:
                    self.pruning_steps += 1
                    node.shift_this_step[:] = False
                    for i, param in enumerate(self.parameters()):
                        param_item = param.item()
                        while param_item > np.pi:
                            param_item -= 2 * np.pi
                        while param_item < -np.pi:
                            param_item += 2 * np.pi
                        self.last_abs_param[i] = 0.01 + np.abs(param_item)
                    idx = torch.multinomial(
                        self.last_abs_param, int(self.sampling_ratio * self.n_params)
                    )
                    node.shift_this_step[idx] = True
                    if self.pruning_steps == self.pruning_window_size:
                        self.is_accumulation = True
                        self.pruning_steps = 0

            self.num_forwards += 1 + 2 * np.sum(node.shift_this_step)
            node_out, time_spent = node.shift_and_run(
                x,
                use_qiskit=use_qiskit,
                is_last_node=(k == self.n_nodes - 1),
                is_first_node=(k == 0),
                parallel=False,
            )
            # logger.info('Time spent:')
            # logger.info(time_spent)
            x = node_out
            mse_all.append(F.mse_loss(node_out, node.x_before_act_quant))
            if verbose:
                acts = {
                    "x_before_add_noise": node.x_before_add_noise.cpu().detach().data,
                    "x_before_norm": node.x_before_norm.cpu().detach().data,
                    "x_before_add_noise_second": node.x_before_add_noise_second.cpu()
                    .detach()
                    .data,
                    "x_before_act_quant": node.x_before_act_quant.cpu().detach().data,
                    "x_after_act_quant": node_out.cpu().detach().data,
                }
                self.activations.append(acts)

        self.mse_all = mse_all

        if verbose:
            os.makedirs(os.path.join(configs.run_dir, "activations"), exist_ok=True)
            torch.save(
                self.activations,
                os.path.join(
                    configs.run_dir, "activations", f"{configs.eval_config_dir}.pt"
                ),
            )
            # logger.info(f"[use_qiskit]={use_qiskit},
            # expectation:\n {x.data}")

        if getattr(self.arch, "output_len", None) is not None:
            x = x.reshape(bsz, -1, self.arch.output_len).sum(-1)

        if x.dim() > 2:
            x = x.squeeze()

        x = F.log_softmax(x, dim=1)
        return x

    def backprop_grad(self):
        for k, node in reversed(list(enumerate(self.nodes))):
            grad_output = node.circuit_out.grad
            for i, param in enumerate(node.q_layer.parameters()):
                if node.shift_this_step[i]:
                    param.grad = (
                        torch.sum(node.grad_qlayer[i] * grad_output)
                        .to(dtype=torch.float32)
                        .view(param.shape)
                    )
                else:
                    self.count1 = self.count1 + 1
                    param.grad = (
                        torch.tensor(0.0)
                        .to(dtype=torch.float32, device=param.device)
                        .view(param.shape)
                    )
                self.last_abs_grad[i] = np.abs(param.grad.item())
                # if (np.abs(param.grad.item()) < 0):
                #     param.grad = torch.tensor(0.).to(dtype=torch.float32, device=param.device).view(param.shape)
                #     self.count1 = self.count1 + 1
                self.count2 = self.count2 + 1

            inputs_grad2loss = None
            for input_grad in node.grad_encoder:
                input_grad2loss = torch.sum(input_grad * grad_output, dim=1).view(-1, 1)
                if inputs_grad2loss is None:
                    inputs_grad2loss = input_grad2loss
                else:
                    inputs_grad2loss = torch.cat((inputs_grad2loss, input_grad2loss), 1)

            if k != 0:
                node.circuit_in.backward(inputs_grad2loss)
        # logger.info(str(self.count1) + '/' + str(self.count2))


model_dict = {
    "q_multifc0": QMultiFCModel0,
}



import torch.nn as nn
import torchquantum as tq

from abc import ABCMeta

__all__ = [
    "QuantumModule",
    "QuantumModuleList",
    "QuantumModuleDict",
]


class QuantumModule(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.static_mode = False
        self.graph = None
        self.parent_graph = None
        self.is_graph_top = False
        self.unitary = None
        self.wires = None
        self.n_wires = None
        self.q_device = None
        # this is for gpu or cpu, not q device
        self.device = None
        # for the static tensor network simulation optimizations
        self.wires_per_block = None
        self.qiskit_processor = None
        self.noise_model_tq = None

    def set_noise_model_tq(self, noise_model_tq):
        for module in self.modules():
            module.noise_model_tq = noise_model_tq

    def set_qiskit_processor(self, processor):
        for module in self.modules():
            module.qiskit_processor = processor

    def set_wires_per_block(self, wires_per_block):
        self.wires_per_block = wires_per_block

    def static_on(self, is_graph_top=True, wires_per_block=3):
        self.wires_per_block = wires_per_block
        # register graph of itself and parent
        self.static_mode = True
        self.is_graph_top = is_graph_top
        if self.graph is None:
            self.graph = tq.QuantumGraph()

        for module in self.children():
            if isinstance(module, nn.ModuleList) or isinstance(module, nn.ModuleDict):
                # if QuantumModuleList or QuantumModuleDict, its graph will
                # be the same as the parent graph because ModuleList and
                # ModuleDict do not call the forward function
                module.graph = self.graph
            module.parent_graph = self.graph
            if not isinstance(module, tq.QuantumDevice):
                module.static_on(is_graph_top=False)

    def static_off(self):
        self.static_mode = False
        self.graph = None
        for module in self.children():
            if not isinstance(module, tq.QuantumDevice):
                module.static_off()

    def set_graph_build_finish(self):
        self.graph.is_list_finish = True
        for module in self.graph.module_list:
            if not isinstance(module, tq.QuantumDevice):
                module.set_graph_build_finish()

    def static_forward(self, q_device: tq.QuantumDevice):
        self.q_device = q_device
        self.device = q_device.states.device
        self.graph.q_device = q_device
        self.graph.device = q_device.states.device
        # self.unitary, self.wires, self.n_wires = \
        self.graph.forward(wires_per_block=self.wires_per_block)
        # tqf.qubitunitary(
        #     q_device=self.q_device,
        #     wires=self.wires,
        #     params=self.unitary
        # )

    def get_unitary(self, q_device: tq.QuantumDevice, x=None):
        original_wires_per_block = self.wires_per_block
        original_static_mode = self.static_mode
        self.static_off()
        self.static_on(wires_per_block=q_device.n_wires)
        self.q_device = q_device
        self.device = q_device.state.device
        self.graph.q_device = q_device
        self.graph.device = q_device.state.device

        self.is_graph_top = False
        # forward to register all modules to the module list, but do not
        # apply the unitary to the state vector
        if x is None:
            self.forward(q_device)
        else:
            self.forward(q_device, x)
        self.is_graph_top = True

        self.graph.build(wires_per_block=q_device.n_wires)
        self.graph.build_static_matrix()
        unitary = self.graph.get_unitary()

        self.static_off()
        if original_static_mode:
            self.static_on(original_wires_per_block)

        return unitary


class QuantumModuleList(nn.ModuleList, QuantumModule, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class QuantumModuleDict(nn.ModuleDict, QuantumModule, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def test():
    pass


if __name__ == "__main__":
    test()



import copy
from typing import Dict, Iterable, List, TYPE_CHECKING

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from opt_einsum import contract
from qiskit import IBMQ
from qiskit.exceptions import QiskitError
from qiskit.providers.aer.noise.device.parameters import gate_error_values
from torchpack.utils.config import Config
from torchpack.utils.logging import logger

import torchquantum as tq
from torchquantum.macro import C_DTYPE


if TYPE_CHECKING:
    from torchquantum.module import QuantumModule
    from torchquantum.devices import QuantumDevice
else:
    QuantumModule = None
    QuantumDevice = None


__all__ = [
    "pauli_eigs",
    "diag",
    "Timer",
    "get_unitary_loss",
    "legalize_unitary",
    "switch_little_big_endian_matrix",
    "switch_little_big_endian_state",
    "get_expectations_from_counts",
    "find_global_phase",
    "build_module_op_list",
    "build_module_from_op_list",
    "build_module_description_test",
    "get_p_v_reg_mapping",
    "get_p_c_reg_mapping",
    "get_v_c_reg_mapping",
    "get_cared_configs",
    "get_success_rate",
    "get_provider",
    "get_provider_hub_group_project",
    "normalize_statevector",
    "get_circ_stats",
    "partial_trace",
    "tensor_form",
    "matrix_form",
    "dm_to_mixture_of_state",
]


def pauli_eigs(n) -> np.ndarray:
    r"""Eigenvalues for :math:`A^{\o times n}`, where :math:`A` is
    Pauli operator, or shares its eigenvalues.

    As an example if n==2, then the eigenvalues of a tensor product consisting
    of two matrices sharing the eigenvalues with Pauli matrices is returned.

    Args:
        n (int): the number of qubits the matrix acts on
    Returns:
        list: the eigenvalues of the specified observable
    """
    if n == 1:
        return np.array([1, -1])
    return np.concatenate([pauli_eigs(n - 1), -pauli_eigs(n - 1)])


def diag(x):
    # input tensor, output tensor with diagonal as the input
    # manual implementation because torch.diag does not support autograd of
    # complex number
    diag_len = x.shape[-1]
    x = x.unsqueeze(-1)
    dims = list(x.shape)
    x = torch.cat([x, torch.zeros(dims[:-1] + [diag_len]).to(x.device)], dim=-1)
    x = x.view(dims[:-2] + [diag_len * (diag_len + 1)])[..., :-diag_len]
    x = x.view(dims[:-2] + [diag_len, diag_len])
    return x


class Timer(object):
    def __init__(self, device="gpu", name="", times=100):
        self.device = device
        self.name = name
        self.times = times
        if device == "gpu":
            self.start = torch.cuda.Event(enable_timing=True)
            self.end = torch.cuda.Event(enable_timing=True)

    def __enter__(self):
        if self.device == "gpu":
            self.start.record()

    def __exit__(self, exc_type, exc_value, tb):
        if self.device == "gpu":
            self.end.record()
            torch.cuda.synchronize()
            print(
                f"Task: {self.name}: "
                f"{self.start.elapsed_time(self.end) / self.times} ms"
            )


def get_unitary_loss(model: nn.Module):
    loss = 0
    for name, params in model.named_parameters():
        if "TrainableUnitary" in name:
            U = params
            like_identity = U.matmul(U.conj().permute(0, 2, 1))
            identity = torch.eye(U.shape[0], dtype=C_DTYPE, device=U.device)
            loss += F.mse_loss(
                torch.view_as_real(identity), torch.view_as_real(like_identity)
            )

    return loss


def legalize_unitary(model: nn.Module):
    with torch.no_grad():
        for name, params in model.named_parameters():
            if "TrainableUnitary" in name:
                U = params
                U, Sigma, V = torch.svd(U)
                params.data.copy_(U.matmul(V.conj().permute(0, 2, 1)))


def switch_little_big_endian_matrix(mat):
    if len(mat.shape) % 2 == 1:
        is_batch_matrix = True
        bsz = mat.shape[0]
        reshape = [bsz] + [2] * int(np.log2(mat[0].size))
    else:
        is_batch_matrix = False
        reshape = [2] * int(np.log2(mat.size))

    original_shape = mat.shape
    mat = mat.reshape(reshape)
    axes = list(range(len(mat.shape) // 2))
    axes.reverse()
    axes += [axis + len(mat.shape) // 2 for axis in axes]

    if is_batch_matrix:
        axes = [0] + [axis + 1 for axis in axes]

    mat = np.transpose(mat, axes=axes).reshape(original_shape)
    return mat


def switch_little_big_endian_state(state):
    if len(state.shape) > 1:
        is_batch_state = True
        bsz = state.shape[0]
        reshape = [bsz] + [2] * int(np.log2(state[0].size))
    elif len(state.shape) == 1:
        is_batch_state = False
        reshape = [2] * int(np.log2(state.size))
    else:
        logger.exception(f"Dimension of statevector should be 1 or 2")
        raise ValueError

    original_shape = state.shape
    state = state.reshape(reshape)

    if is_batch_state:
        axes = list(range(1, len(state.shape)))
        axes.reverse()
        axes = [0] + axes
    else:
        axes = list(range(len(state.shape)))
        axes.reverse()

    mat = np.transpose(state, axes=axes).reshape(original_shape)

    return mat


def switch_little_big_endian_matrix_test():
    logger.info(switch_little_big_endian_matrix(np.ones((16, 16))))
    logger.info(switch_little_big_endian_matrix(np.ones((5, 16, 16))))


def switch_little_big_endian_state_test():
    logger.info(switch_little_big_endian_state(np.ones((5, 16))))
    logger.info(switch_little_big_endian_state(np.arange(8)))


def get_expectations_from_counts(counts, n_wires):
    exps = []
    if isinstance(counts, dict):
        counts = [counts]
    for count in counts:
        ctr_one = [0] * n_wires
        total_shots = 0
        for k, v in count.items():
            for wire in range(n_wires):
                if k[wire] == "1":
                    ctr_one[wire] += v
            total_shots += v
        prob_one = np.array(ctr_one) / total_shots
        exp = np.flip(-1 * prob_one + 1 * (1 - prob_one))
        exps.append(exp)
    res = np.stack(exps)
    return res


def find_global_phase(mat1, mat2, threshold):
    for i in range(mat1.shape[0]):
        for j in range(mat1.shape[1]):
            # find a numerical stable global phase
            if np.abs(mat1[i][j]) > threshold and np.abs(mat2[i][j]) > threshold:
                return mat2[i][j] / mat1[i][j]
    return None


def build_module_op_list(m: QuantumModule, x=None) -> List:
    """
    serialize all operations in the module and generate a list with
    [{'name': RX, 'has_params': True, 'trainable': True, 'wires': [0],
    n_wires: 1, 'params': [array([[0.01]])]}]
    so that an identity module can be reconstructed
    The module needs to have static support
    """

    m.static_off()
    m.static_on(wires_per_block=None)
    m.is_graph_top = False

    # forward to register all modules and parameters
    if x is None:
        m.forward(q_device=None)
    else:
        m.forward(q_device=None, x=x)

    m.is_graph_top = True
    m.graph.build_flat_module_list()

    module_list = m.graph.flat_module_list
    m.static_off()

    op_list = []

    for module in module_list:
        if module.params is not None:
            if module.params.shape[0] > 1:
                # more than one param, so it is from classical input with
                # batch mode
                assert not module.has_params
                params = None
            else:
                # has quantum params, batch has to be 1
                params = module.params[0].data.cpu().numpy()
        else:
            params = None

        op_list.append(
            {
                "name": module.name.lower(),
                "has_params": module.has_params,
                "trainable": module.trainable,
                "wires": module.wires,
                "n_wires": module.n_wires,
                "params": params,
            }
        )

    return op_list


def build_module_from_op_list(
    op_list: List[Dict], remove_ops=False, thres=None
) -> QuantumModule:
    logger.info(f"Building module from op_list...")
    thres = 1e-5 if thres is None else thres
    n_removed_ops = 0
    ops = []
    for info in op_list:
        params = info["params"]

        if remove_ops:
            if params is not None:
                params = (
                    np.array(params)
                    if isinstance(params, Iterable)
                    else np.array([params])
                )
                params = params % (2 * np.pi)
                params[params > np.pi] -= 2 * np.pi
                if all(abs(params) < thres):
                    n_removed_ops += 1
                    continue

        op = tq.op_name_dict[info["name"]](
            has_params=info["has_params"],
            trainable=info["trainable"],
            wires=info["wires"],
            n_wires=info["n_wires"],
            init_params=info["params"],
        )
        ops.append(op)

    if n_removed_ops > 0:
        logger.warning(f"Remove in total {n_removed_ops} pruned operations.")
    else:
        logger.info(f"Do not remove any operations.")

    return tq.QuantumModuleFromOps(ops)


def build_module_description_test():
    import pdb

    from torchquantum.plugins import tq2qiskit

    pdb.set_trace()
    from examples.core.models.q_models import QFCModel12

    q_model = QFCModel12({"n_blocks": 4})
    desc = build_module_op_list(q_model.q_layer)
    print(desc)
    q_dev = tq.QuantumDevice(n_wires=4)
    m = build_module_from_op_list(desc)
    tq2qiskit(q_dev, m, draw=True)

    desc = build_module_op_list(
        tq.RandomLayerAllTypes(n_ops=200, wires=[0, 1, 2, 3], qiskit_compatible=True)
    )
    print(desc)
    m1 = build_module_from_op_list(desc)
    tq2qiskit(q_dev, m1, draw=True)


def get_p_v_reg_mapping(circ):
    """
    p are physical qubits
    v are logical qubits
    """
    p2v_orig = circ._layout.get_physical_bits().copy()
    mapping = {
        "p2v": {},
        "v2p": {},
    }

    for p, v in p2v_orig.items():
        if v.register.name == "q":
            mapping["p2v"][p] = v.index
            mapping["v2p"][v.index] = p

    return mapping


def get_p_c_reg_mapping(circ):
    """
    p are physical qubits
    c are classical registers
    """
    mapping = {
        "p2c": {},
        "c2p": {},
    }
    for gate in circ.data:
        if gate[0].name == "measure":
            mapping["p2c"][gate[1][0].index] = gate[2][0].index
            mapping["c2p"][gate[2][0].index] = gate[1][0].index

    return mapping


def get_v_c_reg_mapping(circ):
    """
    p are physical qubits, the real fabricated qubits
    v are logical qubits, also the 'wires' in torchquantum lib
    c are classical registers
    want to get v2c
    """

    p2v_orig = circ._layout.get_physical_bits().copy()
    p2v = {}
    for p, v in p2v_orig.items():
        if v.register.name == "q":
            p2v[p] = v.index

    mapping = {
        "p2c": {},
        "c2p": {},
    }
    for gate in circ.data:
        if gate[0].name == "measure":
            mapping["p2c"][gate[1][0].index] = gate[2][0].index
            mapping["c2p"][gate[2][0].index] = gate[1][0].index

    mapping2 = {"v2c": {}, "c2v": {}}

    for c, p in mapping["c2p"].items():
        mapping2["c2v"][c] = p2v[p]

    for c, v in mapping2["c2v"].items():
        mapping2["v2c"][v] = c

    return mapping2


def get_cared_configs(conf, mode) -> Config:
    """only preserve cared configs"""
    conf = copy.deepcopy(conf)
    ignores = [
        "callbacks",
        "criterion",
        "debug",
        "legalization",
        "regularization",
        "verbose",
        "get_n_params",
        "prune",
    ]

    if "super" not in conf.trainer.name:
        ignores.append("scheduler")
        ignores.append("optimizer")

    for ignore in ignores:
        if hasattr(conf, ignore):
            delattr(conf, ignore)

    if hasattr(conf, "dataset"):
        dataset_ignores = [
            "binarize",
            "binarize_threshold",
            "center_crop",
            "name",
            "resize",
            "resize_mode",
            "root",
            "train_valid_split_ratio",
        ]
        for dataset_ignore in dataset_ignores:
            if hasattr(conf.dataset, dataset_ignore):
                delattr(conf.dataset, dataset_ignore)

    if not mode == "es" and hasattr(conf, "es"):
        delattr(conf, "es")
    elif mode == "es" and hasattr(conf, "es") and hasattr(conf.es, "eval"):
        delattr(conf.es, "eval")

    if not mode == "train" and hasattr(conf, "trainer"):
        delattr(conf, "trainer")

    if hasattr(conf, "qiskit"):
        qiskit_ignores = [
            "seed_simulator",
            "seed_transpiler",
            "coupling_map_name",
            "basis_gates_name",
            "est_success_rate",
        ]
        for qiskit_ignore in qiskit_ignores:
            if hasattr(conf.qiskit, qiskit_ignore):
                delattr(conf.qiskit, qiskit_ignore)

    if hasattr(conf, "run"):
        run_ignores = ["device", "workers_per_gpu", "n_epochs"]
        for run_ignore in run_ignores:
            if hasattr(conf.run, run_ignore):
                delattr(conf.run, run_ignore)

    return conf


def get_success_rate(properties, transpiled_circ):
    # estimate the success rate according to the error rates of single and
    # two-qubit gates in transpiled circuits

    gate_errors = gate_error_values(properties)
    # construct the error dict
    gate_error_dict = {}
    for gate_error in gate_errors:
        if gate_error[0] not in gate_error_dict.keys():
            gate_error_dict[gate_error[0]] = {tuple(gate_error[1]): gate_error[2]}
        else:
            gate_error_dict[gate_error[0]][tuple(gate_error[1])] = gate_error[2]

    success_rate = 1
    for gate in transpiled_circ.data:
        gate_success_rate = (
            1 - gate_error_dict[gate[0].name][tuple(map(lambda x: x.index, gate[1]))]
        )
        if gate_success_rate == 0:
            gate_success_rate = 1e-5
        success_rate *= gate_success_rate

    return success_rate


def get_provider(backend_name, hub=None):
    # mass-inst-tech-1 or MIT-1
    if backend_name in ["ibmq_casablanca", "ibmq_rome", "ibmq_bogota", "ibmq_jakarta"]:
        if hub == "mass" or hub is None:
            provider = IBMQ.get_provider(
                hub="ibm-q-research", group="mass-inst-tech-1", project="main"
            )
        elif hub == "mit":
            provider = IBMQ.get_provider(
                hub="ibm-q-research", group="MIT-1", project="main"
            )
        else:
            raise ValueError(f"not supported backend {backend_name} in hub " f"{hub}")
    elif backend_name in [
        "ibmq_paris",
        "ibmq_toronto",
        "ibmq_manhattan",
        "ibmq_guadalupe",
        "ibmq_montreal",
    ]:
        provider = IBMQ.get_provider(hub="ibm-q-ornl", group="anl", project="csc428")
    else:
        if hub == "mass" or hub is None:
            try:
                provider = IBMQ.get_provider(
                    hub="ibm-q-research", group="mass-inst-tech-1", project="main"
                )
            except QiskitError:
                # logger.warning(f"Cannot use MIT backend, roll back to open")
                logger.warning(f"Use the open backend")
                provider = IBMQ.get_provider(hub="ibm-q")
        elif hub == "mit":
            provider = IBMQ.get_provider(
                hub="ibm-q-research", group="MIT-1", project="main"
            )
        else:
            provider = IBMQ.get_provider(hub="ibm-q")

    return provider


def get_provider_hub_group_project(hub="ibm-q", group="open", project="main"):
    provider = IBMQ.get_provider(
        hub=hub,
        group=group,
        project=project,
    )
    return provider


def normalize_statevector(states):
    # make sure the square magnitude of statevector sum to 1
    # states = states.contiguous()
    original_shape = states.shape
    states_reshape = states.reshape(states.shape[0], -1)

    # for states with no energy, need to set all zero state as energy 1
    energy = (abs(states_reshape) ** 2).sum(dim=-1)
    if energy.min() == 0:
        for k, val in enumerate(energy):
            if val == 0:
                states_reshape[k][0] = 1

    factors = torch.sqrt(1 / ((abs(states_reshape) ** 2).sum(dim=-1))).unsqueeze(-1)
    states = (states_reshape * factors).reshape(original_shape)

    return states


def get_circ_stats(circ):
    depth = circ.depth()
    width = circ.width()
    size = circ.size()
    n_single_gates = 0
    n_two_gates = 0
    n_three_more_gates = 0
    n_gates_dict = {}
    n_measure = 0

    for gate in circ.data:
        op_name = gate[0].name
        wires = list(map(lambda x: x.index, gate[1]))
        if op_name in n_gates_dict.keys():
            n_gates_dict[op_name] += 1
        else:
            n_gates_dict[op_name] = 1

        if op_name == "measure":
            n_measure += 1
        elif len(wires) == 1:
            n_single_gates += 1
        elif len(wires) == 2:
            n_two_gates += 1
        else:
            n_three_more_gates += 1

    return {
        "depth": depth,
        "size": size,
        "width": width,
        "n_single_gates": n_single_gates,
        "n_two_gates": n_two_gates,
        "n_three_more_gates": n_three_more_gates,
        "n_gates_dict": n_gates_dict,
    }


def partial_trace(
    q_device: QuantumDevice,
    keep_indices: List[int],
) -> torch.Tensor:
    """Returns a density matrix with only some qubits kept.
    Args:
        q_device: The q_device to take the partial trace over.
        keep_indices: Which indices to take the partial trace of the
            state_vector on.
    Returns:
        A density matrix with only the qubits specified by keep_indices.
    """
    n_wires = q_device.n_wires
    keep_indices = np.array(keep_indices) + 1
    dm_left_index = np.arange(n_wires + 1)
    dm_right_index = np.arange(n_wires + 1)
    dm_right_index[keep_indices] *= -1

    dm_out_index = np.array([0])
    dm_out_index = np.concatenate((dm_out_index, keep_indices))
    dm_out_index = np.concatenate((dm_out_index, -keep_indices))

    dm = contract(
        q_device.states,
        dm_left_index,
        q_device.states.conj(),
        dm_right_index,
        dm_out_index,
    )
    return dm


def tensor_form(dm: torch.Tensor):
    """Returns the tensor form of a density matrix.
    Args:
        dm: a (batched) density matrix.
    Returns:
        The tensor form of the density matrix.
    """
    size = dm.size()
    batched = len(size) % 2 == 1
    if batched:
        n_wires = int(np.log2(np.prod(size) / size[0]) / 2)
        shape_list = [size[0]] + [2] * (2 * n_wires)
    else:
        n_wires = int(np.log2(np.prod(size)) / 2)
        shape_list = [2] * (2 * n_wires)
    return dm.reshape(shape_list)


def matrix_form(dm: torch.Tensor):
    """Returns the matrix form of a density matrix.
    Args:
        dm: a (batched) density matrix.
    Returns:
        The matrix form of the density matrix.
    """
    size = dm.size()
    batched = len(size) % 2 == 1
    if batched:
        n_wires = int(np.log2(np.prod(size) / size[0]) / 2)
        shape_list = [size[0]] + [2**n_wires] * 2
    else:
        n_wires = int(np.log2(np.prod(size)) / 2)
        shape_list = [2**n_wires] * 2
    return dm.reshape(shape_list)


def dm_to_mixture_of_state(dm: torch.Tensor, atol=1e-10):
    """
    Args:
        q_device: The state vector to take the partial trace over.
        keep_indices: Which indices to take the partial trace of the
            state_vector on.
        atol: The tolerance for determining that a factored state is pure.
    """
    size = dm.size()
    batched = len(size) % 2 == 1
    if batched:
        dims = int(np.log2(np.prod(size) / size[0]) / 2)

        eigvals, eigvecs = torch.linalg.eigh(dm)
        result_list = []
        for batch in range(size[0]):
            mixture = tuple(
                zip(
                    eigvals[batch],
                    [vec.reshape([2] * dims) for vec in eigvecs[batch].T],
                )
            )
            result_list.append(
                tuple([(float(p[0]), p[1]) for p in mixture if p[0] > 1e-10])
            )
        return result_list
    else:

        dims = int(np.log2(np.prod(size)) / 2)

        eigvals, eigvecs = torch.linalg.eigh(dm)
        mixture = tuple(zip(eigvals, [vec.reshape([2] * dims) for vec in eigvecs.T]))
        return tuple([(float(p[0]), p[1]) for p in mixture if p[0] > 1e-10])


def partial_trace_test():
    import torchquantum.functional as tqf

    n_wires = 4
    q_device = tq.QuantumDevice(n_wires=n_wires)

    tqf.hadamard(q_device, wires=0)
    tqf.cnot(q_device, wires=[0, 1])

    dms = partial_trace(q_device, [0])
    dms = matrix_form(dms)
    mixture = dm_to_mixture_of_state(dms)

    print(mixture)


if __name__ == "__main__":
    build_module_description_test()
    switch_little_big_endian_matrix_test()
    switch_little_big_endian_state_test()
